# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.


from sqlalchemy.sql.expression import and_

from basesoft.util.virtual_date import today
from basesoft.web.contrib.order.controllers import AbstractOrderController
from basesoft.web.contrib.order.models import OrderLineRule
from basesoft.web.core.util.controllers import transaction_decorator

from we.core.controllers import AbstractWeController
from we.declaration.models import WeDeclaration
from we.member.models import WeLpp
from we.order.models import WeDeclarationOrder, WeDeclarationOrderLine
from we.product.models import WeHqppRegulationPeriodProduct
from we.tbs.models import WeTbsLicense


class WeDeclarationOrderController(AbstractWeController, AbstractOrderController):
    _instance_class = WeDeclarationOrder

    def set_additional_fields(self):
        self._set_additional_fields({
            self._instance_class: ['lpp', 'year', 'month', 'quarter', 'bi_annual']
        }, {
            WeDeclarationOrderLine: ['tbs_product_level', 'weight', 'is_b2b'],
            OrderLineRule: ['tax_class_code', 'gl_account_code'],  # FIXME this should be added through parent!
        })

    def _load_state_for_export(self, order):
        rv = super(WeDeclarationOrderController, self)._load_state_for_export(order)
        rv['properties'] = [{
            'key': 'year',
            'value_int': order.year
        },{
            'key': 'month',
            'value_int': order.month
        },{
            'key': 'quarter',
            'value_int': order.quarter
        },{
            'key': 'bi_annual',
            'value_int': order.bi_annual
        },{
            'key': 'producer_relation_id',
            'value_str': order.lpp.lpp_relation_id  # this is the ID the TBS should have in their system
        },{
            'key': 'invoice_relation_id',
            'value_str': order.lpp.invoice_relation_id  # FIXME, is this still necessary?
        }]
        return rv

    def _load_lines_state(self, lines, for_export=False):

        if for_export:
            return self.__load_lines_state_for_export(lines)
        # FIXME!
        return super(WeDeclarationOrderController, self)._load_lines_state(lines)

    def __load_lines_state_for_export(self, lines):
        rv = []

        for line in lines:

            state_line = {
                'index': line.index,
                'polymorphic_type': line.polymorphic_type,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'properties': []
            }

            if line.polymorphic_type == 'WeDeclarationOrderLine':

                state_line['properties'] += [{
                        'key': 'weight',
                        'value_decimal': line.weight,
                    },{
                        'key': 'local_category',
                        'value_str': line.tbs_product_level.label,
                    },{
                        'key': 'local_category_description',
                        'value_str': line.tbs_product_level.description,
                    }]

                if line.is_b2b is not WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE:
                    state_line['properties'].append({
                        'key': 'dimension',
                        'value_str': 'B2B' if line.is_b2b == WeHqppRegulationPeriodProduct.IS_B2B_TRUE else 'B2C'
                    })

            rv.append(state_line)

        return rv

    @transaction_decorator
    def process_declarations(self, tbs_id):
        tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.license_id == tbs_id).one()

        # check the day
        closing_day = tbs.get_config('we.tbs.declaration.closing.day')

        values = self.get_tbs_period_for_day(tbs, closing_day)

        if values is None:
            self.logger.info('Today (%s) is not the closing date for TBS %s' % (today(), tbs))
            return None

        self.logger.info('Declaration closing date reached for TBS %s' % tbs)

        year = values[0]
        biannual = values[1]
        quarter = values[2]
        month = values[3]
        # get all active lpps at this tbs
        lpps = self.sa_session.query(WeLpp).filter(
            and_(
                WeLpp.tbs == tbs,
                WeLpp.status == WeLpp.STATUS_ACTIVE
            )
        )

        rv = []

        for lpp in lpps:
            o = self._process_declarations_for_lpp(lpp, year, biannual, quarter, month)
            if o is not None:
                rv.append(o)
        return rv

    def _process_declarations_for_lpp(self, lpp, year, biannual, quarter, month):
        self.logger.info(' - Processing complete Declarations for LPP %s' % lpp)

        if month is not None:
            months = [month]
        elif quarter is not None:
            months = range(((quarter-1)*3)+1, (quarter*3)+1)
        elif biannual is not None:
            months = range(1, 7) if biannual == WeDeclarationOrder.BI_ANNUAL_FIRST_HALF else range(7, 13)
        else:
            months = range(1, 13)

        # get all unprocessed declarations per lpp,
        declarations = self.sa_session.query(WeDeclaration).filter(
            and_(
                WeDeclaration.lpp == lpp,
                WeDeclaration.status == WeDeclaration.STATUS_COMPLETE,
                WeDeclaration.declaration_order == None,
                WeDeclaration.year == year,
                WeDeclaration.month.in_(months)
            )
        ).order_by(WeDeclaration.month).all()
        if len(declarations) == 0:
            self.logger.info(' - No completed unexported Declarations found.')
            return None

        self.logger.info(' - %d complete Declarations found. Squashing them per Local Category.' % len(declarations))

        lines = {}
        ordered = []
        for declaration in declarations:
            for line in declaration.lines:
                rpp = line.hqpp_period_product.tbs_regulation_period_product
                key = (rpp.id, line.hqpp_period_product.is_b2b)
                if key not in ordered:
                    ordered.append(key)
                if not lines.has_key(key):
                    lines[key] = {'rpp': rpp, 'lines': [], 'is_b2b': line.hqpp_period_product.is_b2b}
                lines[key]['lines'].append(line)

        #order, index, merchant, description, quantity, price, polymorphic_type=OrderLine,
        #apply_vat=True, gl_account_code=None, tax_class=None, dry_run=False, gl_account_number=None, **kwargs

        order_lines = []
        for key in ordered:
            product = lines[key]
            rpp = product['rpp']
            ol = {
                'polymorphic_type': WeDeclarationOrderLine,
                'description': rpp.tbs_product_level.label,
                'tbs_product_level': rpp.tbs_product_level,
                'price': rpp.fee,
                'quantity': 0,
                'weight': 0,
                'is_b2b': product['is_b2b']
            }
            order_lines.append(ol)
            # sum quantity and weight for all lines with same local category
            for line in product['lines']:
                ol['quantity'] += line.quantity
                ol['weight'] += line.product_total_weight

        # save order
        order = self._save_or_update_order(None, order_lines, merchant=lpp.tbs, scope=lpp.hqpp.default_group, lpp=lpp,
                                           year=year, bi_annual=biannual, quarter=quarter, month=month)
        self.logger.info(' - Added new order %s' % order)

        for declaration in declarations:
            declaration.status = WeDeclaration.STATUS_EXPORTED
            declaration.declaration_order = order
            self.sa_session.add(declaration)

        return order
