# Copyright (c) 2016 BaseSoft B.V.
# See LICENSE for details.


import copy
from decimal import Decimal
from datetime import date

import datetime

from basesoft.util.virtual_date import set_date
from basesoft.web.contrib.file.models import ImportFile, ImportFileType, File
from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.core.exceptions import PublicSecurityException
from sqlalchemy.sql.expression import and_

from sqlalchemy.orm.exc import NoResultFound
from we.declaration.controllers import WeDeclarationController
from we.declaration.importers import WeDeclarationXlsxImporter, WeDeclarationAbstractImporter
from we.declaration.models import WeDeclaration
from we.declaration.schedule import WeDeclarationImportTask, WeDeclarationReminderTask
from we.member.controllers import WeMemberController
from we.product.controllers import WeProductController
from we.product.models import WeCN8, WeHqppRegulationPeriodProduct, WeTbsRegulationPeriodProduct
from we.tbs.models import WeTbsRegulationPeriod, WeTbsLicense
from we.test.unit.we_unittest import WeUnitTestCase


class DeclarationTestCase(WeUnitTestCase):

    def setUp(self):
        self.connect_db()

        self._operator_account = self._create_operator_account()
        self.controller = WeDeclarationController(self.sa_session, account_id=self._operator_account.id)

    def test_save_and_load_declaration(self):

        member, lpp, tbs_product_level, tbs = self.create_lpp()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 6, 10))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)
        product2 = self.add_hqpp_product(member.id, pl3.id, code.code, 'SK596853', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]
        period_product2 = product2.get_period_product(rp1)[0]
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        # line 1
        # product weight = 2.123
        # fee = 12.345
        # product total weight = 10.615
        # product_total_fee = 131.04218

        # save new declaration based on weight(default)
        declaration = self.controller.save_or_update_declaration(None, state)
        self.assertEqual(declaration.lpp_id, state['lpp_id'])
        self.assertEqual(len(declaration.lines), len(state['lines']))
        self.assertEqual(declaration.lines[0].quantity, state['lines'][0]['quantity'])
        self.assertEqual(declaration.lines[0].product_total_weight, Decimal('10.61500'))
        self.assertEqual(declaration.lines[0].product_total_fee, Decimal('131.04218'))
        self.assertEqual(declaration.lines[0].product_total_fee, Decimal('131.04218'))
        self.assertEqual(declaration.lines[0].hqpp_period_product, period_product)
        self.assertEqual(declaration.total_fee, Decimal('135.29667'))

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': declaration.lines[0].id,
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 10,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': declaration.lines[1].id,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product1.id
                }, {
                    'line_id': None,
                    'product_id': product2.id,
                    'sku': 'SK596853',
                    'quantity': 3,
                    'hqpp_period_product_id': period_product2.id
                }
            ]
        }
        # update
        declaration = self.controller.save_or_update_declaration(declaration.id, state)
        self.assertEqual(len(declaration.lines), len(state['lines']))
        self.assertEqual(declaration.lines[0].quantity, state['lines'][0]['quantity'])
        self.assertEqual(declaration.lines[0].hqpp_period_product, period_product)
        self.assertEqual(declaration.total_weight, Decimal('38.21400'))
        self.assertEqual(declaration.total_fee, Decimal('276.28722'))

        # try to load declaration for a new HQPP account (should raise exception)
        member, acc = self._create_member(create_account=True)
        self._attach_permissions_to_group(acc.groups[0].group, 'we_view_declaration')

        # instanciate the controller with the HQPP acc
        controller = WeDeclarationController(self.sa_session, account_id=acc.id, check_active=False)
        self.assertRaises(PublicSecurityException, controller.load_declaration, declaration.id)

        # try to load declaration for a new TBS acc (should raise exception)
        tbs = self._create_tbs(self._operator_account.id)
        acc = self._create_active_account(license=self.sa_session.query(License).filter(License.name == tbs.name).first())
        self._attach_permissions_to_group(acc.groups[0].group, 'we_view_declaration')

        # instanciate the controller with the TBS acc
        controller = WeDeclarationController(self.sa_session, account_id=acc.id, check_active=False)
        self.assertRaises(PublicSecurityException, controller.load_declaration, declaration.id)

        # load declaration with the operator acc (always have access)
        state = self.controller.load_declaration(declaration.id)
        self.assertEqual(len(declaration.lines), len(state['lines']))
        self.assertEqual(declaration.lines[0].quantity, state['lines'][0]['quantity'])
        self.assertEqual(declaration.lines[1].quantity, state['lines'][1]['quantity'])

        # delete a line
        del state['lines'][2]
        declaration = self.controller.save_or_update_declaration(declaration.id, state)
        self.assertEqual(len(declaration.lines), len(state['lines']))

    def test_save_declaration_for_sku_at_multiple_product_levels(self):
        """
        Test declaration gets the correct hqpp period product for hqpp product
        that is linked to different local categories in different regulation periods
        :return:
        """
        member, lpp, tbs_product_level, tbs = self.create_lpp()

        # create another product level at different tbs
        product_level, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 2, 10))

        rp = self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 1, [{
            'id': tbs_product_level.id,
            'fee': Decimal('12.345')
        }])

        rp1 = self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 5, [{
            'id': product_level.id,
            'fee': Decimal('0.123')
        }])

        product = self.add_hqpp_product(member.id, tbs_product_level.id, code.code, 'UC389458')
        trpp = self.sa_session.query(WeTbsRegulationPeriodProduct).filter(
            and_(
                WeTbsRegulationPeriodProduct.tbs_regulation_period == rp1,
                WeTbsRegulationPeriodProduct.tbs_product_level == product_level
            )
        ).one()
        set_date(date(2015, 6, 10))
        # update the product details with the new product level
        hrpp = self._add_new_hqpp_period_product(product, trpp, rp1)
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': hrpp.id
                }
            ]
        }

        # there is hqpp period product linked to the regulation period with year 2015 and month 2
        self.assertRaises(NoResultFound, self.controller.save_or_update_declaration, None, state)

        hrpp = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
            and_(
                WeHqppRegulationPeriodProduct.product == product,
                WeHqppRegulationPeriodProduct.tbs_regulation_period == rp
            )
        ).one()
        state['lines'][0]['hqpp_period_product_id'] = hrpp.id
        declaration = self.controller.save_or_update_declaration(None, state)

    def test_save_and_update_declaration_lines_b2b_enabled(self):

        member, lpp, tbs_product_level, tbs = self.create_lpp()

        # enable b2b/b2c setting
        tbs.set_config('we.tbs.require.end.user.distinction', True)

        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        }])

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 6, 10))

        product = self.add_hqpp_product(member.id, [pl1.id, pl1.id], code.code, 'UC389458', is_b2b=True, period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, [pl1.id, pl1.id], code.code, 'WEE34958', is_b2b=True, period=rp1.start_date)

        period_product = product.get_period_product(rp1, True)
        period_product1 = product1.get_period_product(rp1, True)

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product[0].id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1[0].id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1[1].id
                }
            ]
        }

        declaration = self.controller.save_or_update_declaration(None, state)
        self.assertEqual(len(declaration.lines), len(state['lines']))
        self.assertEqual(declaration.lines[0].hqpp_period_product, period_product[0])
        self.assertEqual(declaration.lines[1].hqpp_period_product, period_product1[0])
        self.assertEqual(declaration.lines[2].hqpp_period_product, period_product1[1])

        state = self.controller.load_declaration(declaration.id)
        self.assertEqual(state['lines'][0]['hqpp_period_product_id'], period_product[0].id)
        self.assertEqual(state['lines'][1]['hqpp_period_product_id'], period_product1[0].id)
        self.assertEqual(state['lines'][2]['hqpp_period_product_id'], period_product1[1].id)

        del state['lines'][0]
        dec = self.controller.save_or_update_declaration(declaration.id, state)
        self.assertEqual(len(dec.lines), len(state['lines']))
        self.assertEqual(declaration.lines[0].hqpp_period_product, period_product1[0])
        self.assertEqual(declaration.lines[1].hqpp_period_product, period_product1[1])

    def test_declaration_resubmit_with_new_product(self):
        member, lpp, tbs_product_level, tbs = self.create_lpp()

        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('10.345')
        }])

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 5, 10))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]

        wpc = WeProductController(self.sa_session, account_id=self._operator_account.id)

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 7,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }
            ]
        }
        # save new declaration
        declaration = self.controller.save_or_update_declaration(None, state)
        self.assertEqual(declaration.lines[0].hqpp_period_product, period_product)

        set_date(date(2015, 10, 1))

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 7,
            'lines': [
                {
                    'line_id': declaration.lines[0].id,
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 10,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': '',
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        # resubmit
        declaration = self.controller.save_or_update_declaration(declaration.id, state)
        self.assertEqual(declaration.lines[1].hqpp_period_product, period_product1)

    def test_different_weight_calculations(self):
        member, lpp, tbs_product_level, tbs = self.create_lpp()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        }], weight_settings=WeTbsRegulationPeriod.WEIGHT_PER_UNIT+WeTbsRegulationPeriod.WEIGHT_CATEGORY_1)

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 6, 10))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        # line 1
        # product weight = 2.123
        # category 1 weight = 12.000
        # fee = 12.345

        # save new declaration based on weight(default)
        declaration = self.controller.save_or_update_declaration(None, state)
        self.assertEqual(declaration.lpp_id, state['lpp_id'])
        self.assertEqual(len(declaration.lines), len(state['lines']))
        self.assertEqual(declaration.lines[0].quantity, state['lines'][0]['quantity'])
        self.assertEqual(declaration.lines[0].product_total_weight, Decimal('70.61500'))
        self.assertEqual(declaration.lines[0].product_total_fee, Decimal('871.74218'))
        self.assertEqual(declaration.lines[0].hqpp_period_product, period_product)
        self.assertEqual(declaration.lines[1].product_total_weight, Decimal('28.24600'))
        self.assertEqual(declaration.lines[1].product_total_fee, Decimal('28.30249'))
        self.assertEqual(declaration.lines[1].hqpp_period_product, period_product1)
        self.assertEqual(declaration.total_fee, Decimal('900.04467'))

    def test_validate_declaration_file(self):
        """
        Test the sku validation method.
        :return:
        """
        set_date(date(2015, 6, 10))

        member, lpp, tbs_product_level, tbs = self.create_lpp()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        rp = self.create_tbs_regulation_period(tbs.id, 'Test RP', 2015, 1, [{
            'id': tbs_product_level.id,
            'fee': Decimal('20.12')
        },{
            'id': pl1.id,
            'fee': Decimal('33.33')
        }])

        code = self.sa_session.query(WeCN8).first()

        product = self.add_hqpp_product(member.id, tbs_product_level.id, code.code, 'UC389458')
        product1 = self.add_hqpp_product(member.id, pl1.id, code.code, 'WEE34958')

        headers = ['LPP name', 'LPP_ID', 'Year', 'Month', 'SKU', 'B2B/B2C', 'Quantity']
        row1 = ['garbage', lpp.id, 2015, 3, 'UC389458', '', 5]
        row2 = ['garbage', lpp.id, 2015, 3, 'WEE34958', '', 10]
        row3 = ['garbage', lpp.id, 2015, 3, 'UCKF9458', '', 3]
        row4 = ['garbage', lpp.id, 2015, 3, 'SK596853', '', 3]

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        book.save('test_declaration_xlsx.xlsx')

        f = open('test_declaration_xlsx.xlsx', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)

        # xlsx importer
        importer = WeDeclarationXlsxImporter(self.sa_session, account_id=self._operator_account.id)
        file_path = importer.get_file_path(myfile)

        missing_skus = self.controller.validate_declaration(importer, file_path, member.id)
        self.assertEqual(len(missing_skus), 2)

    def test_declaration_import(self):
        """
        Declaration import is used to import multiple declarations for multiple LPPs in one go.
        """
        set_date(date(2015, 6, 10))

        member, lpp, tbs_product_level, tbs = self.create_lpp()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        # add second LPP
        c = WeMemberController(self.sa_session, account_id=self._operator_account.id)
        lpp_state = self.get_lpp_state(member.id, tbs.id, vat_no='AR00000000000')
        lpp2 = c.save_or_update_company_details(None, lpp_state)

        # add LPP to TBS
        state = {
            'tbs_links': [{
                'lpp_id': lpp2.id,
                'tbs_id': tbs.id,
                'lpp_company_id': lpp2.company.id,
                'ar_company_id': None,
                'invoice_company_id': None
            }]
        }
        c.save_companies_linked_to_tbs(member.id, state)

        code = self.sa_session.query(WeCN8).first()

        # add tbs product levels to regulation period
        self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 1, [{
            'id': tbs_product_level.id,
            'fee': Decimal('0.12')
        },{
            'id': pl1.id,
            'fee': Decimal('122.33')
        }])
        self.add_hqpp_product(member.id, tbs_product_level.id, code.code, 'UC389458')
        self.add_hqpp_product(member.id, pl1.id, code.code, 'WEE34958')

        headers = ['LPP name', 'LPP_ID', 'Year', 'Month', 'SKU', 'B2B/B2C', 'Quantity']
        row1 = ['garbage', lpp.id, 2015, 3, 'UC389458', '', 5]
        row2 = ['garbage', lpp.id, 2015, 3, 'WEE34958', '', 10]
        row3 = ['garbage', lpp2.id, 2015, 6, 'UC389458', '', 3]
        row4 = ['garbage', lpp2.id, 2015, 5, 'WEE34958', '', 3]
        row5 = ['garbage', lpp2.id, 2015, 6, 'WEE34958', '', 3]

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        sheet1.append(row5)
        book.save('test_declaration_xlsx.xlsx')

        f = open('test_declaration_xlsx.xlsx', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)

        # xlsx importer
        importer = WeDeclarationAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        row_count = importer.get_row_count('test_declaration_xlsx.xlsx')
        importer.import_file(myfile, self.controller, hqpp_id=member.id,
                             row_count=row_count, account_id=self._operator_account.id)

        # get the saved file
        f = self.sa_session.query(ImportFile).join(ImportFileType)\
            .filter(and_(ImportFileType.name == importer.IMPORT_FILE_TYPE,
                         ImportFile.status == ImportFile.STATUS_NEW))\
            .order_by(ImportFile.last_updated.desc()).first()
        self.assertEqual(f.file_name, 'test_declaration_xlsx.xlsx')
        self.assertEqual(f.settings, '{"row_count": %s, "hqpp_id": %s, '
                                     '"account_id": %s}' % (row_count, member.id, self._operator_account.id))

        WeDeclarationImportTask().execute()

        d1 = self.sa_session.query(WeDeclaration).filter(WeDeclaration.lpp == lpp).one()
        self.assertEqual(len(d1.lines), 2)
        self.assertEqual(d1.total_quantity, 15)
        self.assertEqual(d1.month, 3)
        self.assertEqual(d1.year, 2015)

        d2 = self.sa_session.query(WeDeclaration).filter(WeDeclaration.lpp == lpp2).all()
        self.assertEqual(len(d2), 2)
        dec21 = self.sa_session.query(WeDeclaration).filter(and_(WeDeclaration.lpp == lpp2,
                                                                 WeDeclaration.year == 2015,
                                                                 WeDeclaration.month == 6)).one()
        self.assertEqual(len(dec21.lines), 2)
        self.assertEqual(dec21.total_quantity, 6)
        db_skus = [line.hqpp_period_product.product.sku for line in dec21.lines]
        self.assertTrue(all(sku in db_skus for sku in ['UC389458', 'WEE34958']))

        # update a declaration
        self.add_hqpp_product(member.id, pl1.id, code.code, 'WEEE123')

        headers = ['LPP name', 'LPP_ID', 'Year', 'Month', 'SKU', 'B2B/B2C', 'Quantity']
        row1 = ['garbage', lpp.id, 2015, 3, 'UC389458', '', 5]
        row2 = ['garbage', lpp.id, 2015, 3, 'WEE34958', '', 10]
        row3 = ['garbage', lpp2.id, 2015, 6, 'UC389458', '', 3]
        row4 = ['garbage', lpp2.id, 2015, 5, 'WEE34958', '', 3]
        row5 = ['garbage', lpp2.id, 2015, 6, 'WEE34958', '', 3]
        row6 = ['garbage', lpp.id, 2015, 3, 'WEEE123', '', 5]

        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        sheet1.append(row5)
        sheet1.append(row6)
        book.save('test_declaration_xlsx.xlsx')

        f = open('test_declaration_xlsx.xlsx', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)
        row_count = importer.get_row_count('test_declaration_xlsx.xlsx')
        importer.import_file(myfile, self.controller, hqpp_id=member.id,
                             row_count=row_count, account_id=self._operator_account.id)

        WeDeclarationImportTask().execute()

        d1 = self.sa_session.query(WeDeclaration).filter(WeDeclaration.lpp == lpp).one()
        self.assertEqual(len(d1.lines), 3)
        self.assertEqual(d1.total_quantity, 20)
        db_skus = [line.hqpp_period_product.product.sku for line in d1.lines]
        self.assertTrue(all(sku in db_skus for sku in ['UC389458', 'WEE34958', 'WEEE123']))

    def test_declaration_import_with_b2b(self):
        """
        Declaration import is used to import multiple declarations for multiple LPPs in one go.
        """
        set_date(date(2015, 6, 10))

        member, lpp, tbs_product_level, tbs = self.create_lpp()
        tbs.set_config('we.tbs.require.end.user.distinction', True)
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        # add second LPP
        c = WeMemberController(self.sa_session, account_id=self._operator_account.id)
        lpp_state = self.get_lpp_state(member.id, tbs.id, vat_no='AR00000000000')
        lpp2 = c.save_or_update_company_details(None, lpp_state)

        # add LPP to TBS
        state = {
            'tbs_links': [{
                'lpp_id': lpp2.id,
                'tbs_id': tbs.id,
                'lpp_company_id': lpp2.company.id,
                'ar_company_id': None,
                'invoice_company_id': None
            }]
        }
        c.save_companies_linked_to_tbs(member.id, state)

        code = self.sa_session.query(WeCN8).first()

        # add tbs product levels to regulation period
        self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 1, [{
            'id': tbs_product_level.id,
            'fee': Decimal('0.12')
        },{
            'id': pl1.id,
            'fee': Decimal('122.33')
        }])
        self.add_hqpp_product(member.id, [tbs_product_level.id, pl1.id], code.code, 'UC389458', is_b2b=True)
        self.add_hqpp_product(member.id, [pl1.id, tbs_product_level.id], code.code, 'WEE34958', is_b2b=True)

        headers = ['LPP name', 'LPP_ID', 'Year', 'Month', 'SKU', 'B2B/B2C', 'Quantity']
        row1 = ['garbage', lpp.id, 2015, 3, 'UC389458', 'B2B', 5]
        row2 = ['garbage', lpp.id, 2015, 3, 'WEE34958', 'B2C', 10]
        row3 = ['garbage', lpp2.id, 2015, 6, 'UC389458', 'B2b', 3]
        row5 = ['garbage', lpp2.id, 2015, 6, 'WEE34958', '', 3] # error

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row5)
        book.save('test_declaration_xlsx.xlsx')

        f = open('test_declaration_xlsx.xlsx', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)

        # xlsx importer
        importer = WeDeclarationAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        row_count = importer.get_row_count('test_declaration_xlsx.xlsx')
        importer.import_file(myfile, self.controller, hqpp_id=member.id,
                             row_count=row_count, account_id=self._operator_account.id)

        WeDeclarationImportTask().execute()

        # get the failures file
        f = self.sa_session.query(File).filter(File.file_name == 'failures_test_declaration_xlsx.csv')\
            .order_by(ImportFile.last_updated.desc()).first()
        importer = WeDeclarationAbstractImporter()\
            .get_importer(f)(self.sa_session, account_id=self._operator_account.id)
        file_path = importer.get_file_path(f)
        self.assertEqual(importer.get_row_count(file_path), 2) # with headers

        d1 = self.sa_session.query(WeDeclaration).filter(WeDeclaration.lpp == lpp).one()
        self.assertEqual(len(d1.lines), 2)
        self.assertEqual(d1.total_quantity, 15)
        self.assertEqual(d1.month, 3)
        self.assertEqual(d1.year, 2015)

        is_b2b = [l.hqpp_period_product.is_b2b for l in d1.lines]
        self.assertTrue(all(b in is_b2b for b in [WeHqppRegulationPeriodProduct.IS_B2B_FALSE,
                                                  WeHqppRegulationPeriodProduct.IS_B2B_TRUE]))

        # check notification is created
        nc = NotificationController(self.sa_session, account_id=self._operator_account.id)
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 1)
        self.assertEqual(notif[0].callback, '{"js_module": "we_declaration_manager"}')

    def test_declaration_import_with_exported_declaration(self):
        """
        Declaration import is used to import multiple declarations for multiple LPPs in one go.
        """
        set_date(date(2015, 6, 10))

        member, lpp, tbs_product_level, tbs = self.create_lpp()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        code = self.sa_session.query(WeCN8).first()

        # add tbs product levels to regulation period
        rp = self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 1, [{
            'id': tbs_product_level.id,
            'fee': Decimal('0.12')
        },{
            'id': pl1.id,
            'fee': Decimal('122.33')
        }])
        product = self.add_hqpp_product(member.id, tbs_product_level.id, code.code, 'UC389458')
        product2 = self.add_hqpp_product(member.id, pl1.id, code.code, 'WEE34958')

        # create exported declaration
        period_product = product.get_period_product(rp)[0]
        period_product2 = product2.get_period_product(rp)[0]
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 6,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product2.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product2.id
                }
            ]
        }

        # save new declaration based on weight(default)
        declaration = self.controller.save_or_update_declaration(None, state)
        declaration.export_date = date(2015, 3, 20)
        self.sa_session.add(declaration)
        self.sa_session.flush()

        headers = ['LPP name', 'LPP_ID', 'Year', 'Month', 'SKU', 'B2B/B2C', 'Quantity']
        row1 = ['garbage', lpp.id, 2015, 3, 'UC389458', '', 5]
        row2 = ['garbage', lpp.id, 2015, 3, 'WEE34958', '', 10]
        row3 = ['garbage', lpp.id, 2015, 6, 'UC389458', '', 3]
        row5 = ['garbage', lpp.id, 2015, 6, 'WEE34958', '', 3]

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row5)
        book.save('test_declaration_xlsx.xlsx')

        f = open('test_declaration_xlsx.xlsx', 'r')
        data = f.read()
        f.close()

        # validate
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)
        importer = WeDeclarationXlsxImporter(self.sa_session, account_id=self._operator_account.id)
        file_path = importer.get_file_path(myfile)
        validate_rv = self.controller.validate_declaration(importer, file_path, member.id)

        self.assertEqual(validate_rv[0], 'Declaration for year 2015 and month 6 has been exported to the TBS. Cannot update.')

    def test_declaration_import_empty_lines(self):
        """
        test for #445
        """
        set_date(date(2015, 6, 10))

        member, lpp, tbs_product_level, tbs = self.create_lpp()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        # add second LPP
        c = WeMemberController(self.sa_session, account_id=self._operator_account.id)
        lpp_state = self.get_lpp_state(member.id, tbs.id, vat_no='AR00000000000')
        lpp2 = c.save_or_update_company_details(None, lpp_state)

        # add LPP to TBS
        state = {
            'tbs_links': [{
                'lpp_id': lpp2.id,
                'tbs_id': tbs.id,
                'lpp_company_id': lpp2.company.id,
                'ar_company_id': None,
                'invoice_company_id': None
            }]
        }
        c.save_companies_linked_to_tbs(member.id, state)

        code = self.sa_session.query(WeCN8).first()

        # add tbs product levels to regulation period
        self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 1, [{
            'id': tbs_product_level.id,
            'fee': Decimal('0.12')
        },{
            'id': pl1.id,
            'fee': Decimal('122.33')
        }])
        self.add_hqpp_product(member.id, tbs_product_level.id, code.code, 'UC389458')
        self.add_hqpp_product(member.id, pl1.id, code.code, 'WEE34958')

        headers = ['LPP name', 'LPP_ID', 'Year', 'Month', 'SKU', 'B2B/B2C', 'Quantity']
        row1 = ['garbage', lpp.id, 2015, 3, 'UC389458', '', 5]
        row2 = ['garbage', lpp.id, 2015, 3, 'WEE34958', '', 10]
        row3 = ['garbage', lpp2.id, 2015, 6, 'UC389458', '', 3]
        row4 = ['garbage', lpp2.id, 2015, 5, 'WEE34958', '', 3]
        row6 = [' ', ' ', ' ', ' ', ' ', ' ', ' ']
        row5 = ['garbage', lpp2.id, 2015, 6, 'WEE34958', '', 3]

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        sheet1.append(row5)
        sheet1.append(row6)
        book.save('test_declaration_xlsx.xlsx')

        f = open('test_declaration_xlsx.xlsx', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)

        # validate, used to go wrong in #445
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)
        importer = WeDeclarationXlsxImporter(self.sa_session, account_id=self._operator_account.id)
        file_path = importer.get_file_path(myfile)
        missing_skus = self.controller.validate_declaration(importer, file_path, member.id)

        # check that the file can still be imported properly
        importer = WeDeclarationAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        row_count = importer.get_row_count('test_declaration_xlsx.xlsx')

        print 'start import'
        importer.import_file(myfile, self.controller, hqpp_id=member.id,
                             row_count=row_count, account_id=self._operator_account.id)
        print 'end import'

        # get the saved file
        f = self.sa_session.query(ImportFile).join(ImportFileType)\
            .filter(and_(ImportFileType.name == importer.IMPORT_FILE_TYPE,
                         ImportFile.status == ImportFile.STATUS_NEW))\
            .order_by(ImportFile.last_updated.desc()).first()
        self.assertEqual(f.file_name, 'test_declaration_xlsx.xlsx')
        self.assertEqual(f.settings, '{"row_count": %s, "hqpp_id": %s, '
                                     '"account_id": %s}' % (row_count, member.id, self._operator_account.id))

        WeDeclarationImportTask().execute()

        d1 = self.sa_session.query(WeDeclaration).filter(WeDeclaration.lpp == lpp).one()
        self.assertEqual(len(d1.lines), 2)
        self.assertEqual(d1.total_quantity, 15)
        self.assertEqual(d1.month, 3)
        self.assertEqual(d1.year, 2015)

        d2 = self.sa_session.query(WeDeclaration).filter(WeDeclaration.lpp == lpp2).all()
        self.assertEqual(len(d2), 2)
        dec21 = self.sa_session.query(WeDeclaration).filter(and_(WeDeclaration.lpp == lpp2,
                                                                 WeDeclaration.year == 2015,
                                                                 WeDeclaration.month == 6)).one()
        self.assertEqual(len(dec21.lines), 2)
        self.assertEqual(dec21.total_quantity, 6)
        db_skus = [line.hqpp_period_product.product.sku for line in dec21.lines]
        self.assertTrue(all(sku in db_skus for sku in ['UC389458', 'WEE34958']))

        # update a declaration
        self.add_hqpp_product(member.id, pl1.id, code.code, 'WEEE123')

        headers = ['LPP name', 'LPP_ID', 'Year', 'Month', 'SKU', 'B2B/B2C', 'Quantity']
        row1 = ['garbage', lpp.id, 2015, 3, 'UC389458', '', 5]
        row2 = ['garbage', lpp.id, 2015, 3, 'WEE34958', '', 10]
        row3 = ['garbage', lpp2.id, 2015, 6, 'UC389458', '', 3]
        row4 = ['garbage', lpp2.id, 2015, 5, 'WEE34958', '', 3]
        row5 = ['garbage', lpp2.id, 2015, 6, 'WEE34958', '', 3]
        row6 = ['garbage', lpp.id, 2015, 3, 'WEEE123', '', 5]

        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        sheet1.append(row5)
        sheet1.append(row6)
        book.save('test_declaration_xlsx.xlsx')

        f = open('test_declaration_xlsx.xlsx', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_declaration_xlsx.xlsx', data)
        row_count = importer.get_row_count('test_declaration_xlsx.xlsx')
        importer.import_file(myfile, self.controller, hqpp_id=member.id,
                             row_count=row_count, account_id=self._operator_account.id)

        WeDeclarationImportTask().execute()

        d1 = self.sa_session.query(WeDeclaration).filter(WeDeclaration.lpp == lpp).one()
        self.assertEqual(len(d1.lines), 3)
        self.assertEqual(d1.total_quantity, 20)
        db_skus = [line.hqpp_period_product.product.sku for line in d1.lines]
        self.assertTrue(all(sku in db_skus for sku in ['UC389458', 'WEE34958', 'WEEE123']))


    def test_export_declaration(self):
        # create two declarations for different lpps
        member, lpps, tbses = self.create_lpps(self.get_vat_numbers(), lpp_count=2, same_tbs=True)
        tbs = tbses.pop()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 6, 10))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)
        product2 = self.add_hqpp_product(member.id, pl3.id, code.code, 'SK596853', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]
        period_product2 = product2.get_period_product(rp1)[0]
        state = {
            'lpp_id': lpps[0].id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }

        state1 = copy.deepcopy(state)
        state1['lpp_id'] = lpps[1].id
        state1['lines'][0]['sku'] = 'SK596853'
        state1['lines'][0]['product_id'] = product2.id
        state1['lines'][0]['hqpp_period_product_id'] = period_product2.id
        declaration = self.controller.save_or_update_declaration(None, state)
        declaration1 = self.controller.save_or_update_declaration(None, state1)

        # export to file
        _file_name = self.controller.export_declaration(member.id, 2015, 4)

        from openpyxl.reader.excel import load_workbook
        wb = load_workbook(filename=_file_name, use_iterators=True)
        sheet = wb.get_active_sheet()
        row_count = 0
        skus = []
        for row in sheet.iter_rows():
            row_count += 1
            skus.append(row[4].internal_value)

        self.assertEqual(5, row_count)
        self.assertTrue(all(s in skus for s in ['UC389458', 'WEE34958', 'SK596853']))

    def test_export_declaration_new_lpp(self):
        """
        Test that when an LPP has not declared before the created file contains all hqpp products.
        """
        # create two declarations for different lpps
        member, lpps, tbses = self.create_lpps(self.get_vat_numbers(), lpp_count=2, same_tbs=True)
        tbs = tbses.pop()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 6, 10))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)
        product2 = self.add_hqpp_product(member.id, pl3.id, code.code, 'SK596853', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]
        period_product2 = product2.get_period_product(rp1)[0]
        state = {
            'lpp_id': lpps[0].id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }

        declaration = self.controller.save_or_update_declaration(None, state)

        # export to file
        _file_name = self.controller.export_declaration(member.id, 2015, 4)

        row_count = 0
        skus = []

        from openpyxl.reader.excel import load_workbook
        wb = load_workbook(filename=_file_name, use_iterators=True)
        sheet = wb.get_active_sheet()

        for row in sheet.iter_rows():
            row_count += 1
            if row_count > 3:
                skus.append(row[4].internal_value)
                self.assertEqual(row[6].internal_value, None)

        self.assertEqual(6, row_count)
        self.assertTrue(all(s in skus for s in ['UC389458', 'WEE34958', 'SK596853']))

    def test_export_declaration_new_lpps_with_b2b(self):
        """
        Test that when an LPP has not declared before the created file contains all hqpp products.
        """
        # create two declarations for different lpps
        member, lpps, tbses = self.create_lpps(self.get_vat_numbers(), lpp_count=2, same_tbs=True)
        tbs = tbses.pop()
        tbs.set_config('we.tbs.require.end.user.distinction', True)

        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 6, 10))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)
        product2 = self.add_hqpp_product(member.id, pl3.id, code.code, 'SK596853', period=rp1.start_date)

        # export to file
        _file_name = self.controller.export_declaration(member.id, 2015, 4)

        row_count = 0
        skus = []

        from openpyxl.reader.excel import load_workbook
        wb = load_workbook(filename=_file_name, use_iterators=True)
        sheet = wb.get_active_sheet()

        for row in sheet.iter_rows():
            row_count += 1
            if row_count == 1:
                continue
            self.assertTrue(row[5].internal_value in ['B2B', 'B2C'])
        self.assertEqual(13, row_count)

    def test_export_declaration_half_hqpp_products(self):
        """
        Test that only products that are linked to a local category of the tbs of the lpp
        are exported.
        """
        # create two declarations for different lpps
        member, lpp, pl, tbs = self.create_lpp()
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        }])

        rp2 = self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP'), 2015, 5, [{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])
        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 6, 10))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)
        product2 = self.add_hqpp_product(member.id, pl3.id, code.code, 'SK596853', period=rp2.start_date)

        # export to file
        _file_name = self.controller.export_declaration(member.id, 2015, 3)

        from openpyxl.reader.excel import load_workbook
        wb = load_workbook(filename=_file_name, use_iterators=True)
        sheet = wb.get_active_sheet()
        row_count = 0
        skus = []
        for row in sheet.iter_rows():
            row_count += 1
            if row_count == 1:
                continue
            skus.append(row[4].internal_value)
            self.assertEqual(row[2].internal_value, 2015)
            self.assertEqual(row[3].internal_value, 3)

        l = ['UC389458', 'WEE34958']
        self.assertEqual(3, row_count)
        self.assertEqual(len(skus), len(l))
        self.assertTrue(all(s in skus for s in l))

    def test_declaration_reminder_monthly_notification(self):
        # set some date
        set_date(date(2015, 3, 1))

        member, lpps, tbs = self.create_lpps(self.get_vat_numbers(), lpp_count=2, same_tbs=True)
        member, acc = self._create_member(self._operator_account.id, member.id, create_account=True)
        tbs = tbs.pop()
        # run declaration reminder task and check there is a notification created
        WeDeclarationReminderTask().execute()

        # there should be no notification because it is not the day
        nc = NotificationController(self.sa_session, account_id=acc.id)
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 0)

        # set day to reminder day - 7th day of the month for monthly declaring
        reminder_day = 9
        set_date(date(2015, 3, reminder_day))

        # check we have notification
        WeDeclarationReminderTask().execute()
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 1)

        # create declaration for one lpp for month 3
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        }])

        code = self.sa_session.query(WeCN8).first()

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]

        lpp = lpps[0]
        lpp1 = lpps[1]
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 3,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        # save new declaration based on weight(default)
        declaration = self.controller.save_or_update_declaration(None, state)

        # set date to month 4
        set_date(date(2015, 4, reminder_day))
        # check there is another notification for the lpp that doesn't have declaration
        WeDeclarationReminderTask().execute()

        notif = nc.get_unread_notifications()
        # we have 2 unread now because one of the lpp has no declaration
        self.assertEqual(len(notif), 2)

    def test_declaration_reminder_quarterly_notification(self):
        # set some date
        set_date(date(2015, 3, 7))

        member, lpps, tbs = self.create_lpps(self.get_vat_numbers(), lpp_count=2, same_tbs=True)
        member, acc = self._create_member(self._operator_account.id, member.id, create_account=True)
        tbs = tbs.pop()
        # set tbs periodicity to quarterly
        tbs.set_config('we.tbs.declaration.period', WeTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        # run declaration reminder task and check there is a notification created
        WeDeclarationReminderTask().execute()

        # there should be no notification because it is not the day
        nc = NotificationController(self.sa_session, account_id=acc.id)
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 0)

        # set day to reminder day - 7th day of the month for monthly declaring
        reminder_day = 89 - 1 # closing day for quarterly is 89
        tbs.set_config('we.tbs.declaration.reminder.day', reminder_day)
        td = datetime.timedelta(days=reminder_day-1)
        set_date(date(2015, 1, 1) +td)

        # check we have notification
        WeDeclarationReminderTask().execute()
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 1)

        # create declarations for lpps for month 4
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        }])

        code = self.sa_session.query(WeCN8).first()

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]

        lpp = lpps[0]
        lpp1 = lpps[1]
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 4,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        # save new declaration based on weight(default)
        declaration = self.controller.save_or_update_declaration(None, state)

        state['lpp_id'] = lpp1.id
        declaration = self.controller.save_or_update_declaration(None, state)

        # set date to second quarter reminder day
        set_date(date(2015, 7, 1) + td)

        # check there are no new notifcations
        WeDeclarationReminderTask().execute()

        notif = nc.get_unread_notifications()
        # we have 1 unread now because both lpp have declaration in the second quarter
        self.assertEqual(len(notif), 1)

    def test_declaration_reminder_yearly_notification(self):
        # set some date
        set_date(date(2015, 1, 5))

        member, lpps, tbs = self.create_lpps(self.get_vat_numbers(), lpp_count=2, same_tbs=True)
        member, acc = self._create_member(self._operator_account.id, member.id, create_account=True)
        tbs = tbs.pop()
        # set tbs periodicity to yearly
        tbs.set_config('we.tbs.declaration.period', WeTbsLicense.DECLARATION_PERIOD_YEARLY)
        # run declaration reminder task and check there is a notification created
        WeDeclarationReminderTask().execute()

        # there should be no notification because it is not the day
        nc = NotificationController(self.sa_session, account_id=acc.id)
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 0)

        # set day to reminder day - 7th day of the month for monthly declaring
        reminder_day = 364 - 1 # closing day for yearly is 364
        tbs.set_config('we.tbs.declaration.reminder.day', reminder_day)
        td = datetime.timedelta(days=reminder_day-1)
        set_date(date(2014, 1, 1) +td)

        # check we have notification
        WeDeclarationReminderTask().execute()
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 1)

    def test_get_tbs_period_for_day(self):
        tbs = self._create_tbs(self._operator_account)
        tbs.set_config('we.tbs.declaration.period', WeTbsLicense.DECLARATION_PERIOD_MONTHLY)
        set_date(date(2015, 1, 6))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 5), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 7), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 6), None)
        set_date(date(2015, 8, 6))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 5), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 7), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 6), None)

        tbs.set_config('we.tbs.declaration.period', WeTbsLicense.DECLARATION_PERIOD_BIANNUALLY)
        set_date(date(2015, 2, 23))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 53), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 55), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 54), None)
        set_date(date(2015, 8, 23))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 53), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 55), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 54), None)

        tbs.set_config('we.tbs.declaration.period', WeTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        set_date(date(2015, 4, 23))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 112), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 114), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 113), None)
        set_date(date(2015, 8, 23))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 234), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 236), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 235), None)

        tbs.set_config('we.tbs.declaration.period', WeTbsLicense.DECLARATION_PERIOD_YEARLY)
        set_date(date(2015, 2, 23))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 53), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 55), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 54), None)
        set_date(date(2016, 2, 23))
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 53), None)
        self.assertEqual(self.controller.get_tbs_period_for_day(tbs, 55), None)
        self.assertNotEqual(self.controller.get_tbs_period_for_day(tbs, 54), None)

    def _add_new_hqpp_period_product(self, product, trpp, rp):
            hrpp = WeHqppRegulationPeriodProduct()
            hrpp.product = product
            hrpp.tbs_regulation_period_product = trpp
            hrpp.tbs_regulation_period = rp
            self.sa_session.add(hrpp)
            self.sa_session.flush()
            return hrpp