'''
Created on Jan 14, 2015

@author: marten
'''
from datetime import date
from twisted.trial.unittest import SkipTest
from basesoft.web.core.exceptions import PermissionDeniedException, PublicUserException
from basesoft.web.contrib.contact.models import Company
from sqlalchemy.sql.expression import and_
from we.core.controllers import AbstractWeController
from we.tbs.controllers import WeTbsController

from we.test.unit.we_unittest import WeUnitTestCase
from we.member.controllers import WeMemberController
from we.member.models import WeHqppLicense, WeLpp
from basesoft.web.contrib.account.controllers import AccountController
from basesoft.web.contrib.account.models import AccountActivation, Account

class TestCase(WeUnitTestCase):

    def setUp(self):
        self.connect_db()

        self._operator_account = self._create_operator_account()
        self.controller = WeMemberController(self.sa_session, account_id=self._operator_account.id)

    def test_save_and_update_member_by_operator(self):

        # create a new member
        member = self.controller.save_or_update_member(_id=None, state=self._get_member_state('a'))
        member = self.controller.get_member(member.id)
        self.assertEqual(member.role.name, 'WE Producer')

        # verify
        self.assertTrue(isinstance(member, WeHqppLicense))

        # update the member
        state = self.controller.load_member(member.id)
        state['company']['name'] = 'updated company name'

        self.controller.save_or_update_member(member.id, state)

        # verify that the master data change is automatically applied ( operator has approval of master data change )

        member = self.controller.get_member(member.id)
        self.assertEqual(member.temp_company, None)
        self.assertEqual(member.company.name, 'updated company name') # this should also be applied as license name
        self.assertEqual(member.name, 'updated company name')

    def test_create_member_without_account_and_lpp(self):

        master_data = self._get_member_state('a')
        member = self.controller.register_member(master_data, None)
        self.assertEqual(member.company.name, master_data['company']['name'])
        self.assertEqual(len(member.default_group.accounts), 0)

    def test_create_member_with_account(self):

        master_data = self._get_member_state('a')
        account_data = {
            'person': {
                'is_same_person_as': False,
                'last_name': 'Member Contact Last Name',
                'first_name': 'Member Contact First Name',
                'email': 'test-member@mail-test.basesoft.com',
                'preferred_locale': '1',
                'gender': 'male',
                'phone': '234234235',
                'mobile_phone': '123412345',
            },
            'password': {
                'send_link': True
            },
            'account_details': {
                'account_name': self._uniquestring('a')
            }
        }
        # test new person
        member = self.controller.register_member(master_data, account_data)

        self.assertEqual(member.company.name, master_data['company']['name'])

        # verify we have an account
        self.assertEqual(len(member.default_group.accounts), 1)
        account = member.default_group.accounts[0]

        # verify persons are not equal
        self.assertNotEqual(account.person.id, member.company.contact_person.id)

        # verify an account activation link has been added for mailing.
        ac = AccountController(self.sa_session)
        aa = ac.get_account_activation(account, type=AccountActivation.ACTIVATION_TYPE_NEW_ACCOUNT)
        self.assertTrue(isinstance(aa, AccountActivation))

        # now we test another member, but we set the password and we use the same person as company contact.
        account_name = self._uniquestring('a')
        account_data['account_details']['account_name'] = account_name
        account_data['person']['is_same_person_as'] = True
        account_data['password'] = {
            'send_link': False,
            'new_password': 'abcd123456',
            'new_password_verify': 'abcd123456',
        }

        member = self.controller.register_member(master_data, account_data)
        self.assertEqual(len(member.default_group.accounts), 1)

        account = member.default_group.accounts[0]
        self.assertEqual(account.person.id, member.company.contact_person.id)

        account = ac.validate_account(account_name, 'abcd123456')
        self.assertTrue(isinstance(account, Account))

    def test_create_member_with_lpp(self):

        tbs = self._create_tbs(self._operator_account.id)
        master_data = self._get_member_state('a')
        member = self.controller.register_member(master_data, None)

        lpp_state = {
            'lpp': {
                'tbses': [{
                    'tbs_id': tbs.id,
                    'lpp_relation_id': 123,
                    'invoice_relation_id': 123,
                    'lpp_id': None,
                    'use_as_ar': False
                }],
                'member_id': member.id
            },
            'address_data': {'city': u'Boston', 'state': u'Indiana',
                             'street': u'West street', 'country_iso_3166_a2': u'AO',
                             'postcode': u'1132WD', 'country': 7L, 'country_name': u'Angola',
                             'street_2': None, 'country_iso_3166_a3': u'AGO'},
            'contact_person_data': {
                             'first_name': u'John', 'last_name': u'Doe', 'middle_name': u'van', 'polymorphic_type': u'Person',
                             'skype_name': u'BaronJohn', 'title': u'Baron', 'fax': u'123323456', 'phone': u'123456678',
                             'mobile_phone': u'003132435655', 'gender': u'male', 'birth_date': date(1978, 12, 12), 'position': u'Manager',
                             'department': u'IT', 'email': u'doe@john.com', 'preferred_locale': 1
            },
            'bank_data': {
                             'bank_name': u'BankCorp', 'beneficiary': u'What is this?',
                             'number': u'3838383838', 'bic': u'BIC123457',
                             'note': u'Its a note', 'iban': u'AO22BC390594389'
            },
            'website': u'http://wwww.basesoft.com/',
            'name': u'you down with lpp',
            'vat_no': u'RO1630615123457',
        }

        self.controller.save_or_update_company_details(None, lpp_state)

        self.assertEqual(member.company.name, master_data['company']['name'])
        self.assertEqual(len(member.default_group.accounts), 0)
        self.assertEqual(len(member.lpps), 1)
        self.assertEqual(member.lpps[0].lpp_relation_id, '123')
        self.assertEqual(member.lpps[0].invoice_relation_id, '123')

        self.assertEqual(member.lpps[0].company.name, 'you down with lpp')
        self.assertEqual(member.lpps[0].company.addresses[0].address.city, 'Boston')
        self.assertEqual(member.lpps[0].company.contact_person.first_name, 'John')

    def test_update_member_by_member(self):
        """
        Test for updating the master data by the member itself (master data change request)
        """

        member, member_account = self._create_member(create_account=True)
        controller = WeMemberController(self.sa_session, account_id=member_account.id)

        self.assertRaises(PermissionDeniedException, controller.save_or_update_member, _id=None, state=self._get_member_state('a'))

        state = controller.load_member(member.id)
        state['company']['name'] = 'updated company name'

        controller.save_or_update_member(member.id, state)

        # the member should now have a flag set that there is a master data change request.
        member = self.controller.get_member(member.id)
        self.assertTrue(isinstance(member.temp_company, Company))

        self.controller.approve_master_data_change(member.id)

        member = self.controller.get_member(member.id)
        self.assertEqual(member.temp_company, None)
        self.assertEqual(member.company.name, 'updated company name') # this should also be applied as license name
        self.assertEqual(member.name, 'updated company name')

    def test_delete_local_producer_partner(self):
        raise SkipTest('Currently LPPs are enabled on save for easier testing. Remove the skip test when it is over.')
        member_account = self._create_member(create_account=True)
        member = member_account.groups[0].group.license

        tbs = self._create_tbs(self._operator_account.id)

        state = {
            'name': 'My Company Name',
            'lpp': {
                'tbs_id': tbs.id,
                'has_authorized_representative': False
            },
            'vat_no': 'MYVAT'
        }
        lpp = self.controller.add_lpp_to_hqpp(member.id, state)
        lpp = self.sa_session.query(WeLpp).filter(WeLpp.id==lpp.id).first()

        self.controller.delete_lpp(lpp.id)
        lpp = self.sa_session.query(WeLpp).filter(WeLpp.id==lpp.id).first()
        self.assertEqual(lpp, None)

        # test that an active LPP cannot be deleted.
        lpp = self.controller.add_lpp_to_hqpp(member.id, state)
        lpp = self.sa_session.query(WeLpp).filter(WeLpp.id==lpp.id).first()
        lpp.status = WeLpp.STATUS_ACTIVE
        self.sa_session.add(lpp)
        self.sa_session.flush()

        self.assertRaises(PublicUserException, self.controller.delete_lpp, lpp.id)

    def test_deactivate_local_producer_partner(self):
        # we first need the TBS interaction for this.
        pass

    def test_add_local_producer_partners_to_member(self):
        tbs = self._create_tbs(self._operator_account.id)

        master_data = self._get_member_state('a')
        member = self.controller.register_member(master_data, None)

        lpp_state1 = {
            'lpp': {
                'tbses': [{
                    'tbs_id': tbs.id,
                    'lpp_relation_id': 123,
                    'invoice_relation_id': 123,
                    'lpp_id': None,
                    'use_as_ar': False
                }],
                'member_id': member.id
            },
            'address_data': {'city': u'Boston', 'state': u'Indiana',
                             'street': u'West street', 'country_iso_3166_a2': u'AO',
                             'postcode': u'1132WD', 'country': 7L, 'country_name': u'Angola',
                             'street_2': None, 'country_iso_3166_a3': u'AGO'},
            'contact_person_data': {
                             'first_name': u'John', 'last_name': u'Doe', 'middle_name': u'van', 'polymorphic_type': u'Person',
                             'skype_name': u'BaronJohn', 'title': u'Baron', 'fax': u'123323456', 'phone': u'123456678',
                             'mobile_phone': u'003132435655', 'gender': u'male', 'birth_date': date(1978, 12, 12), 'position': u'Manager',
                             'department': u'IT', 'email': u'doe@john.com', 'preferred_locale': 1
            },
            'bank_data': {
                             'bank_name': u'BankCorp', 'beneficiary': u'What is this?',
                             'number': u'3838383838', 'bic': u'BIC123457',
                             'note': u'Its a note', 'iban': u'AO22BC390594389'
            },
            'website': u'http://wwww.basesoft.com/',
            'name': u'you down with lpp',
            'vat_no': u'RO1630615123457',
        }

        lpp_state2 = {
            'lpp': {
                'tbses': [{
                    'tbs_id': tbs.id,
                    'lpp_relation_id': 123,
                    'invoice_relation_id': 123,
                    'lpp_id': None,
                    'use_as_ar': False
                }],
                'member_id': member.id
            },
            'address_data': {'city': u'Houston', 'state': u'Indiana',
                             'street': u'West street', 'country_iso_3166_a2': u'AO',
                             'postcode': u'1132WD', 'country': 7L, 'country_name': u'Angola',
                             'street_2': None, 'country_iso_3166_a3': u'AGO'},
            'contact_person_data': {
                             'first_name': u'Petar', 'last_name': u'Doe', 'middle_name': u'van', 'polymorphic_type': u'Person',
                             'skype_name': u'BaronJohn', 'title': u'Baron', 'fax': u'123323456', 'phone': u'123456678',
                             'mobile_phone': u'003132435655', 'gender': u'male', 'birth_date': date(1978, 12, 12), 'position': u'Manager',
                             'department': u'IT', 'email': u'doe@john.com', 'preferred_locale': 1
            },
            'bank_data': {
                             'bank_name': u'WEEEBANK', 'beneficiary': u'What is this?',
                             'number': u'3838383838', 'bic': u'BIC123457',
                             'note': u'Its a note', 'iban': u'AO22BC390594389'
            },
            'website': u'http://wwww.basesoft.com/',
            'name': u'WEEE comp',
            'vat_no': u'ATU12345620',
        }

        self.controller.save_or_update_company_details(None, lpp_state1)
        self.controller.save_or_update_company_details(None, lpp_state2)

        self.assertEqual(member.company.name, master_data['company']['name'])
        self.assertEqual(len(member.default_group.accounts), 0)
        self.assertEqual(len(member.lpps), 2)

        self.assertEqual(member.lpps[0].company.name, 'you down with lpp')
        self.assertEqual(member.lpps[0].company.addresses[0].address.city, 'Boston')
        self.assertEqual(member.lpps[0].company.contact_person.first_name, 'John')

        self.assertEqual(member.lpps[1].company.name, 'WEEE comp')
        self.assertEqual(member.lpps[1].company.addresses[0].address.city, 'Houston')
        self.assertEqual(member.lpps[1].company.contact_person.first_name, 'Petar')

    def test_get_lpps(self):
        """
        Test for getting all LPPs for operator and own LPPs for HQPP
        """
        # clear WeLpp to make sure there are no other LPPs
        self.sa_session.execute("""
                UPDATE tblWeLpp SET status = %s
        """ % WeLpp.STATUS_INACTIVE)
        # create two HQPPs with LPPs
        member, lpp, local_cat, tbs = self.create_lpp(member_account=True)
        member1, lpp1, local_cat1, tbs1 = self.create_lpp()

        # get the products sheet as an operator = all LPPs in the system
        lpps = self.controller.get_local_producer_partners()
        self.assertEqual(len(lpps), 2)

        # get LPPs for a member account
        self.c = AbstractWeController(self.sa_session, account_id=member.default_group.accounts[0].id)
        lpps = self.c.get_local_producer_partners()

        self.assertEqual(member.id, lpps[0].hqpp_id)
        self.assertEqual(lpp.id, lpps[0].id)
        self.assertEqual(len(member.lpps), len(lpps))

    def test_add_lpp_to_multiple_tbs_same_country(self):
        tc = WeTbsController(self.sa_session, account_id=self._operator_account.id)
        # save first TBS
        state = self._get_tbs_state('TBS')
        vats = self.get_vat_numbers()
        state['vat_no'] = vats[3]
        tbs = tc.save_or_update_tbs(None, state)

        # save second TBS
        state = self._get_tbs_state('TBS')
        vats = self.get_vat_numbers()
        state['vat_no'] = vats[3]
        tbs1 = tc.save_or_update_tbs(None, state)

        # save HQPP and LPP at two TBSes
        master_data = self._get_member_state('a')
        member = self.controller.register_member(master_data, None)

        lpp_state = self.get_lpp_state(member.id, tbs.id)
        lpp = self.controller.save_or_update_company_details(None, lpp_state)

        # add LPP to TBSes
        tbs_links = self.get_lpp_to_tbs_state(lpp, [tbs, tbs1])
        self.controller.save_companies_linked_to_tbs(member.id, tbs_links)

        self.assertEqual(member.lpps[0].tbs, tbs)
        self.assertEqual(member.lpps[1].tbs, tbs1)
        self.assertEqual(member.lpps[0].company, member.lpps[1].company)

    def test_member_is_lpp(self):
        tbs = self._create_tbs(self._operator_account.id)

        master_data = self._get_member_state('HQPP')
        member = self.controller.register_member(master_data, None)

        tbs_links = self.get_lpp_to_tbs_state(member, [tbs], is_member=True)
        self.controller.save_companies_linked_to_tbs(member.id, tbs_links)

        # get saved links TBS => LPP
        lpps = self.sa_session.query(WeLpp).filter(and_(WeLpp.tbs_id != None, WeLpp.hqpp == member)).all()
        self.assertEqual(member.company, lpps[0].company)

        # add more
        tbs1 = self._create_tbs(self._operator_account.id)

        state = {
            'tbs_links': [{
                'lpp_id': lpps[0].id,
                'tbs_id': tbs.id,
                'lpp_company_id': member.company.id,
                'ar_company_id': None,
                'invoice_company_id': None
            }, {
                'lpp_id': None,
                'tbs_id': tbs1.id,
                'lpp_company_id': member.company.id,
                'ar_company_id': None,
                'invoice_company_id': None
            }]
        }
        self.controller.save_companies_linked_to_tbs(member.id, state)

        lpps = self.sa_session.query(WeLpp).filter(and_(WeLpp.tbs_id != None, WeLpp.hqpp == member)).all()
        self.assertEqual(member.company, lpps[0].company)
        self.assertEqual(member.company, lpps[1].company)
        self.assertEqual(tbs, lpps[0].tbs)
        self.assertEqual(tbs1, lpps[1].tbs)

    def test_save_and_update_third_party_company(self):
        # get company state
        state = self.get_third_party_company_state()
        state['polymorphic_type'] = 'WeCompany'
        # save company
        company = self.controller.save_or_update_company_details(None, state)

        self.assertEqual(company.polymorphic_type, 'WeCompany')
        self.assertEqual(company.name, state['name'])

        # update company
        state['name'] = 'WeCorp'
        company = self.controller.save_or_update_company_details(company.id, state)
        self.assertEqual(company.name, state['name'])