# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from datetime import date

from decimal import Decimal

from basesoft.util.virtual_date import set_date, today
from we.declaration.controllers import WeDeclarationController
from we.order.controllers import WeDeclarationOrderController
from we.order.models import WeDeclarationOrder
from we.product.models import WeCN8
from we.tbs.models import WeTbsLicense
from we.test.unit.we_unittest import WeUnitTestCase


class TestCase(WeUnitTestCase):

    def setUp(self):
        self.connect_db()

        self._operator_account = self._create_operator_account()
        self.controller = WeDeclarationOrderController(self.sa_session, account_id=self._operator_account.id)
        set_date(date(2015, 1, 1))

    def _prepare_declarations(self, periodicity):
        controller = WeDeclarationController(self.sa_session, account_id=self._operator_account.id)

        member, lpp, tbs_product_level, tbs = self.create_lpp()

        tbs.set_config('we.tbs.declaration.period', periodicity)
        tbs.set_config('we.tbs.declaration.closing.day', 15)

        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])

        code = self.sa_session.query(WeCN8).first()

        set_date(date(2015, 4, 1))

        product = self.add_hqpp_product(member.id, pl1.id, code.code, 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, code.code, 'WEE34958', period=rp1.start_date)
        product2 = self.add_hqpp_product(member.id, pl3.id, code.code, 'SK596853', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        declaration1 = controller.save_or_update_declaration(None, state)

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 3,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 12,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 8,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        declaration2 = controller.save_or_update_declaration(None, state)

        # Save a Declaration NOT in this Declaration period, it should not be included in the export.

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 5,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 16,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 18,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        declaration3 = controller.save_or_update_declaration(None, state)

        return tbs, (declaration1, declaration2, declaration3,), (period_product, period_product1,)

    def test_create_declaration_order_quarterly(self):
        """
        Creates 2 declarations, aggregates them into an order.
        """
        tbs, declarations, products = self._prepare_declarations(WeTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        tbs_account = self._create_active_account(tbs)

        set_date(date(2015, 4, 15))

        declaration1, declaration2, declaration3 = declarations

        wod = WeDeclarationOrderController(self.sa_session)
        orders = wod.process_declarations(tbs.id)
        self.assertEqual(len(orders), 1)
        order = orders.pop()
        self.assertEqual(order.year, today().year)
        self.assertEqual(order.quarter, 1)
        self.assertEqual(order.status, WeDeclarationOrder.STATUS_SUBMITTED)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('219.885'))
        self.assertEqual(len(order.lines), 2)

        l = order.lines[0]
        if l.tbs_product_level == products[0].tbs_regulation_period_product.tbs_product_level:
            line1 = l
            line2 = order.lines[1]
        else:
            line1 = order.lines[1]
            line2 = l

        self.assertEqual(line1.quantity, 17)
        self.assertEqual(line1.weight, Decimal('36.091'))
        self.assertEqual(line2.quantity, 10)
        self.assertEqual(line2.weight, Decimal('21.23'))

        self.assertEqual(declaration1.declaration_order, order)
        self.assertEqual(declaration2.declaration_order, order)
        self.assertEqual(declaration3.declaration_order, None)

        wod = WeDeclarationOrderController(self.sa_session, account_id=tbs_account.id)
        wod.assert_privileges_on_object(order, WeDeclarationOrder.PRIV_EXPORT)

        self._attach_permissions_to_group(tbs_account.default_group, 'we_view_declaration_order')

        orders = wod.get_unexported_orders_for_merchant(tbs.id)
        self.assertEqual(len(orders), 1)

    def test_create_declaration_order_monthly(self):
        tbs, declarations, products = self._prepare_declarations(WeTbsLicense.DECLARATION_PERIOD_MONTHLY)

        set_date(date(2015, 3, 15))

        declaration1, declaration2, declaration3 = declarations

        wod = WeDeclarationOrderController(self.sa_session)
        orders = wod.process_declarations(tbs.id)
        self.assertEqual(len(orders), 1)
        order = orders.pop()

        self.assertEqual(order.year, today().year)
        self.assertEqual(order.quarter, None)
        self.assertEqual(order.month, 2)
        self.assertEqual(order.status, WeDeclarationOrder.STATUS_SUBMITTED)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('63.729'))
        self.assertEqual(len(order.lines), 2)

        self.assertEqual(declaration1.declaration_order, order)
        self.assertEqual(declaration2.declaration_order, None)
        self.assertEqual(declaration3.declaration_order, None)

        # advance a month
        set_date(date(2015, 4, 15))

        wod = WeDeclarationOrderController(self.sa_session)
        orders = wod.process_declarations(tbs.id)
        self.assertEqual(len(orders), 1)
        order = orders.pop()

        self.assertEqual(order.year, today().year)
        self.assertEqual(order.quarter, None)
        self.assertEqual(order.month, 3)
        self.assertEqual(order.status, WeDeclarationOrder.STATUS_SUBMITTED)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('156.156'))
        self.assertEqual(len(order.lines), 2)

        self.assertNotEqual(declaration1.declaration_order, None)
        self.assertNotEqual(declaration1.declaration_order, order)
        self.assertEqual(declaration2.declaration_order, order)
        self.assertEqual(declaration3.declaration_order, None)

    def test_create_declaration_order_skip_month(self):
        # This can happen if a member doesnt do their declarations until after a period is closed.
        # FIXME, what should happen (take this along with Resubmits)?
        tbs, declarations, products = self._prepare_declarations(WeTbsLicense.DECLARATION_PERIOD_MONTHLY)

        set_date(date(2015, 4, 15))

        declaration1, declaration2, declaration3 = declarations

        wod = WeDeclarationOrderController(self.sa_session)
        orders = wod.process_declarations(tbs.id)
        self.assertEqual(len(orders), 1)
        order = orders.pop()

        self.assertEqual(declaration1.declaration_order, None)
        self.assertEqual(declaration2.declaration_order, order)
        self.assertEqual(declaration3.declaration_order, None)

    def test_create_declaration_order_yearly(self):
        tbs, declarations, products = self._prepare_declarations(WeTbsLicense.DECLARATION_PERIOD_YEARLY)

        set_date(date(2015, 3, 15))

        declaration1, declaration2, declaration3 = declarations

        wod = WeDeclarationOrderController(self.sa_session)
        orders = wod.process_declarations(tbs.id)
        self.assertEqual(orders, None)

        set_date(date(2016, 1, 15))

        orders = wod.process_declarations(tbs.id)
        order = orders.pop()

        self.assertEqual(order.year, 2015)
        self.assertEqual(order.quarter, None)
        self.assertEqual(order.month, None)
        self.assertEqual(order.status, WeDeclarationOrder.STATUS_SUBMITTED)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('435.441'))
        self.assertEqual(len(order.lines), 2)

        l = order.lines[0]
        if l.tbs_product_level == products[0].tbs_regulation_period_product.tbs_product_level:
            line1 = l
            line2 = order.lines[1]
        else:
            line1 = order.lines[1]
            line2 = l

        self.assertEqual(line1.quantity, 33)
        self.assertEqual(line1.weight, Decimal('70.059'))
        self.assertEqual(line2.quantity, 28)
        self.assertEqual(line2.weight, Decimal('59.444'))

        self.assertEqual(declaration1.declaration_order, order)
        self.assertEqual(declaration2.declaration_order, order)
        self.assertEqual(declaration3.declaration_order, order)
