'''
Created on Dec 22, 2014

@author: rado
'''


import os

from django.utils import simplejson
from sqlalchemy.sql.expression import and_

from basesoft.web.contrib.file.models import ImportFile, ImportFileType
from basesoft.web.contrib.notification.controllers import NotificationController
from we.core.controllers import WeNotificationController
from we.product.importers import HqppProductCsvImporter, HqppProductXlsxImporter, \
    TbsPriceSheetXlsxImporter, TbsPriceSheetCsvImporter, HqppProductAbstractImporter, \
    CN8ToTbsProductLevelAbstractImporter, TbsPriceSheetAbstractImporter, CN8AbstractImporter
from we.product.schedule import HqppProductSheetTask, CN8ToWeeeCategoryTask, CN8ToTbsProductLevelTask, TbsPriceSheetImportTask, \
    RegulationPeriodChangeTask
from we.test.unit.we_unittest import WeUnitTestCase
from we.product.controllers import WeProductController
from we.product.models import WeCN8, WeTbsProductLevel, WeHqppProduct, \
    WeTbsRegulationPeriodProduct, WeHqppRegulationPeriodProduct, WeRegulationPeriodNotification

from decimal import Decimal
from basesoft.web.core.exceptions import PublicSecurityException, PublicUserException
from basesoft.util.virtual_date import set_date
from datetime import date


class TestCase(WeUnitTestCase):

    def setUp(self):
        self.connect_db()

        self._operator_account = self._create_operator_account()
        self.controller = WeProductController(self.sa_session, account_id=self._operator_account.id)
        set_date(date(2015, 8, 8))

    def test_save_or_update_cn8_subcategory(self):

        # remove any links between CN8 and weee category
        self.sa_session.execute("""
                UPDATE tblWeCN8
                SET weee_product_level_id = NULL
        """)
        state = {
            'cn8_codes': [49051000, 63011000, 73211110],
            'weee_product_level_id': '10',  # comes from grouped select, value is subcategory ID
            'unu_code': 9  # use id
        }

        # save new match of cn8 and subcategory
        match = self.controller.save_or_update_cn8_weee_product_level(None, state)
        codes_list = self.controller.get_cn8_by_product_level(match.id)
        self.assertEqual(codes_list[0].unu_code_id, 9)

        codes_list = [cn8_code.code for cn8_code in codes_list]
        # assert that state cn8_codes exist in the returned object
        self.assertTrue(all(value in codes_list for value in state['cn8_codes']))
        self.assertEqual(len(codes_list), len(state['cn8_codes']))
        self.assertEqual(match.id, int(state['weee_product_level_id']))

        # change same codes to different category
        state1 = {
            'cn8_codes': [49051000, 63011000, 73211110],
            'weee_product_level_id': '22',
            'unu_code': 9
        }
        match1 = self.controller.save_or_update_cn8_weee_product_level(None, state1)
        codes_list1 = self.controller.get_cn8_by_product_level(match1.id)
        codes_list1 = [cn8_code.code for cn8_code in codes_list1]
        # assert that state cn8_codes exist in the returned object and subcategory.id is the same
        self.assertTrue(all(value in codes_list1 for value in state1['cn8_codes']))
        self.assertEqual(len(codes_list1), len(state1['cn8_codes']))
        self.assertEqual(match1.id, int(state1['weee_product_level_id']))

        # remove old codes and add more
        state2 = {
            'cn8_codes': [73211190, 73211900, 84041000, 84021200, 84022000],
            'weee_product_level_id': '22',
            'unu_code': 9
        }
        match2 = self.controller.save_or_update_cn8_weee_product_level(None, state2)
        codes_list2 = self.controller.get_cn8_by_product_level(match2.id)
        codes_list2 = [cn8_code.code for cn8_code in codes_list2]
        # assert that state cn8_codes exist in the returned object
        self.assertTrue(all(value in codes_list2 for value in state2['cn8_codes']))
        self.assertEqual(len(codes_list2), 5)
        self.assertEqual(match2.id, int(state2['weee_product_level_id']))

        # test loading as well
        state = self.controller.load_cn8_weee_product_level(match2.id)
        self.assertEqual(state['weee_product_level_id'], match2.id)
        self.assertTrue(all(value in codes_list2 for value in state['cn8_codes']))
        self.assertEqual(len(codes_list2), len(state['cn8_codes']))

    def test_save_or_update_tbs_product_level(self):

        tbs = self._create_tbs(self._operator_account.id)

        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        rp = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP', 2015, 1, 1, [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        },{
            'id': product_level2.id,
            'fee': Decimal('33.33')
        },{
            'id': product_level3.id,
            'fee': Decimal('12.45')
        }])

        product_level = self.controller\
            .save_or_update_tbs_product_level(None, 'TBS-level', 'WEEE', tbs.id,
                                              invoice_basis=WeTbsProductLevel.INVOICE_BASED_ON_QUANTITY,
                                              tbs_regulation_period_id=rp.id,
                                              fee=Decimal("12.123"))

        self.assertEqual(product_level.parent_level, None)
        self.assertEqual(product_level.invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_QUANTITY)
        self.assertEqual(product_level.label, 'TBS-level')

        # save a sublevel
        sublevel = self.controller\
            .save_or_update_tbs_product_level(None, 'TBS-sublevel', 'WEEE sublevel', tbs.id,
                                              parent_level_id=product_level.id,
                                              invoice_basis=WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT,
                                              tbs_regulation_period_id=rp.id,
                                              fee=Decimal("2.123"))

        self.assertEqual(sublevel.parent_level, product_level)
        self.assertEqual(sublevel.invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT)
        self.assertEqual(sublevel.label, 'TBS-sublevel')

        # save a sublevel
        l = self.controller\
            .save_or_update_tbs_product_level(None, 'TBS-sublevel1', 'WEEE sublevel', tbs.id,
                                              parent_level_id=sublevel.id,
                                              invoice_basis=WeTbsProductLevel.INVOICE_BASED_ON_QUANTITY,
                                              tbs_regulation_period_id=rp.id,
                                              fee=Decimal("0.123"))

        self.assertEqual(l.parent_level, sublevel)
        self.assertEqual(l.invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_QUANTITY)
        self.assertEqual(l.label, 'TBS-sublevel1')

        # save a sublevel
        s = self.controller\
            .save_or_update_tbs_product_level(None, 'TBS-sublevel2', 'WEEE sublevel', tbs.id,
                                              parent_level_id=l.id,
                                              invoice_basis=WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT,
                                              tbs_regulation_period_id=rp.id,
                                              fee=Decimal("122.123"))

        self.assertEqual(s.parent_level, l)
        self.assertEqual(s.invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT)
        self.assertEqual(s.label, 'TBS-sublevel2')

        # assert all levels are saved for the regulation period
        rpp = self.sa_session.query(WeTbsRegulationPeriodProduct)\
            .filter(WeTbsRegulationPeriodProduct.tbs_regulation_period_id == rp.id).all()
        self.assertEqual(len(rpp), 7)

    def test_tbs_product_level_archive(self):
        tbs = self._create_tbs(self._operator_account.id)
        tbs1 = self._create_tbs(self._operator_account.id)
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl11, tbs1 = self.create_tbs_product_level(self._operator_account.id, tbs1)

        rp = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP', 2015, 1, 1, [{
            'id': pl1.id,
            'fee': Decimal('20.12')
        }])

        self.assertRaises(PublicUserException, self.controller.mark_tbs_product_levels_as_archived,
                          'TbsProductLevel', [pl1.id, pl11.id])
        l = self.controller.mark_tbs_product_levels_as_archived(None, [pl11.id])
        self.sa_session.expire(pl11)
        self.assertEqual(pl11.archived, 1)

    def test_change_parent_product_level(self):
        """
        Test that loops are not allowed.
        * Product B has Product C as parent.
        * Product C has Product D as parent.
        * Product D has Product B as parent.
        :return:
        """
        tbs = self._create_tbs(self._operator_account.id)

        # create levels
        b, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        c, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        d, tbs = self.create_tbs_product_level(self._operator_account.id, tbs, None, parent_level_id=b.id)
        self.assertRaises(PublicUserException, self.create_tbs_product_level, self._operator_account.id, tbs, b.id, c.id)

    def test_change_product_level_position(self):
        tbs = self._create_tbs(self._operator_account.id)

        # create levels
        b, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        c, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        b.position = 2
        c.position = 1

        b = self.controller.save_or_update_tbs_product_level(b.id, b.label, b.description, tbs.id, position=b.position)
        c = self.controller.save_or_update_tbs_product_level(c.id, c.label, c.description, tbs.id, position=c.position)
        self.assertEqual(c.position, 1)
        self.assertEqual(b.position, 2)

    def test_save_and_load_hqpp_product(self):
        cn8 = self.sa_session.query(WeCN8).first()
        member, lpp, tbs_product_level, tbs = self.create_lpp(member_account=True)

        state = self.get_product_state(member.id, cn8.code, tbs_product_level.id)

        # create tbs regulation period
        self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP_'), 2015, 1,
                                          [{'id': tbs_product_level.id, 'fee': Decimal('2.22')}])
        set_date(date(2015, 8, 1))
        # save product
        product = self.controller.save_or_update_hqpp_product(state['product_id'], state)
        tbs_product_links = self.sa_session.query(WeHqppRegulationPeriodProduct)\
            .filter(WeHqppRegulationPeriodProduct.product_id == product.id).all()

        self.assertEqual(product.sku, state['sku'])
        self.assertEqual(tbs_product_links[0].tbs_regulation_period_product.tbs_product_level_id, tbs_product_level.id)
        self.assertEqual(product.hqpp, member)
        self.assertEqual(product.hs, None)

        # load the saved product
        p = self.controller.load_hqpp_product(product.id)
        self.assertEqual(p['sku'], product.sku)
        self.assertEqual(p['cn8_codes'][0], product.cn8)
        self.assertEqual(p['hqpp_id'], member.id)

        # save another product for another HQPP and check that the new HQPP cannot load a product of the other HQPP
        cn8s = self.sa_session.query(WeCN8).all()
        member1, lpp1, pl1, tbs1 = self.create_lpp(member_account=True)

        state1 = self.get_product_state(member1.id, cn8s[5].code, pl1.id)

        # create tbs regulation period
        self.create_tbs_regulation_period(tbs1.id, self._uniquestring('RP_'), 2015, 1,
                                          [{'id': pl1.id, 'fee': Decimal('11.22')}])

        c = WeProductController(sa_session=self.sa_session, account_id=member1.default_group.accounts[0].id)
        product1 = c.save_or_update_hqpp_product(state1['product_id'], state1)
        self.assertEqual(product1.hqpp_id, member1.id)
        self.assertRaises(PublicSecurityException, c.load_hqpp_product, product.id)

        # disable a link to tbs product level
        product_tbs_state = self.controller.get_hqpp_product_tbs_links(member.id, product.id)
        product_tbs_state[0]['tbs_product_level_id'] = ''
        state['tbs_rowcontrol'] = product_tbs_state

        product = self.controller.save_or_update_hqpp_product(product.id, state)

        set_date(date(2015, 12, 1))

        product_tbs_state = self.controller.get_hqpp_product_tbs_links(member.id, product.id)
        self.assertEqual(product_tbs_state[0].has_key('tbs_product_level_id'), False)

    def test_save_hqpp_product_with_b2b(self):

        cn8 = self.sa_session.query(WeCN8).first()
        member, lpp, tbs_product_level, tbs = self.create_lpp(member_account=True)
        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        state = self.get_product_state(member.id, cn8.code, tbs_product_level.id)

        # create tbs regulation period
        self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP_'), 2015, 1,
                                          [{'id': tbs_product_level.id, 'fee': Decimal('2.22')},
                                           {'id': pl1.id, 'fee': Decimal('0.11')}])
        set_date(date(2015, 8, 1))
        # save product
        product = self.controller.save_or_update_hqpp_product(state['product_id'], state)
        tbs_product_links = self.sa_session.query(WeHqppRegulationPeriodProduct)\
            .filter(WeHqppRegulationPeriodProduct.product_id == product.id).all()

        # we have only 1 record for no B2B
        self.assertEqual(len(tbs_product_links), 1)
        self.assertEqual(tbs_product_links[0].is_b2b, WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE)

        # enable B2B/B2C setting
        tbs.set_config('we.tbs.require.end.user.distinction', True)

        # get product state with B2B
        state = self.get_product_state(member.id, cn8.code, [tbs_product_level.id, pl1.id], is_b2b=True)

        # update product
        product = self.controller.save_or_update_hqpp_product(product.id, state)

        tbs_product_links = self.sa_session.query(WeHqppRegulationPeriodProduct)\
            .filter(WeHqppRegulationPeriodProduct.product_id == product.id).all()
        # we should have 3 lines in the db
        self.assertEqual(len(tbs_product_links), 3)

        # disable one of the B2B/B2C lines
        state_links = self.controller.get_hqpp_product_tbs_links(member.id, product.id)
        state_links[1]['tbs_product_level_id'] = ''
        state['tbs_rowcontrol'] = state_links

        # update product
        product = self.controller.save_or_update_hqpp_product(product.id, state)

        set_date(date(2015, 12, 12))
        product_tbs = self.controller.get_hqpp_product_tbs_links(member.id, product.id)

        # we should have only 1 line B2B/B2C with a TBS product level set
        self.assertEqual(len(product_tbs), 2)
        self.assertEqual(product_tbs[0].has_key('tbs_product_level_id'), True)
        self.assertEqual(product_tbs[1].has_key('tbs_product_level_id'), False)

    def test_update_product_local_category_same_month(self):
        cn8 = self.sa_session.query(WeCN8).first()
        member, lpp, tbs_product_level, tbs = self.create_lpp(member_account=True)
        pl, tbs = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11010')
        state = self.get_product_state(member.id, cn8.code, tbs_product_level.id)

        # create tbs regulation period
        self.create_tbs_regulation_period(tbs.id, self._uniquestring('RP_'), 2015, 1,
                                          [{'id': tbs_product_level.id, 'fee': Decimal('2.22')},
                                           {'id': pl.id, 'fee': Decimal('0.11')}])
        set_date(date(2015, 8, 1))

        # save product
        product = self.controller.save_or_update_hqpp_product(state['product_id'], state)
        tbs_product_links = self.sa_session.query(WeHqppRegulationPeriodProduct)\
            .filter(WeHqppRegulationPeriodProduct.product_id == product.id).all()
        old_product_level_id = tbs_product_links[0].tbs_regulation_period_product.tbs_product_level_id

        # change tbs category for same start_date which should delete the first record and add new one
        product_tbs_state = self.controller.get_hqpp_product_tbs_links(member.id, product.id)
        product_tbs_state[0]['tbs_product_level_id'] = {'all_categories': pl.id}
        state['tbs_rowcontrol'] = product_tbs_state

        product = self.controller.save_or_update_hqpp_product(product.id, state)
        tbs_product_links = self.sa_session.query(WeHqppRegulationPeriodProduct)\
            .filter(WeHqppRegulationPeriodProduct.product_id == product.id).all()
        self.assertEqual(len(tbs_product_links), 1)
        self.assertNotEqual(old_product_level_id, tbs_product_links[0].tbs_regulation_period_product.tbs_product_level_id)

    def test_import_product_sheet(self):
        set_date(date(2015, 5, 5))
        f_name, member, tbses, rps = self.create_product_sheet_csv_file()
        local_cats = ['11010', '3.1', '11030', '3.2']

        # create dict with tbs_id: regulation_period_id
        d = {
            'tbs-%s' % tbses[0].id: rps[0].id,
            'tbs-%s' % tbses[1].id: rps[1].id,
            'tbs-%s' % tbses[2].id: rps[2].id,
            'tbs-%s' % tbses[3].id: rps[3].id
        }
        # csv importer
        importer = HqppProductCsvImporter(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        rv, failures = importer.import_file(f_name, self.controller, hqpp_id=member.id, regulation_periods=d)
        self.assertEqual(len(failures), 0)

        # get product for sku
        product = self.sa_session.query(WeHqppProduct).filter(and_(WeHqppProduct.sku == '210-344',
                                                                   WeHqppProduct.hqpp == member)).one()

        # get linked tbs category
        links = self.sa_session.query(WeHqppRegulationPeriodProduct)\
            .filter(WeHqppRegulationPeriodProduct.product_id == product.id).all()

        # there are 4 links, for two TBS's with a B2B/B2C split
        self.assertEqual(len(links), 4)

        tbs_cats = [c.tbs_regulation_period_product.tbs_product_level.label for c in links]
        self.assertTrue(all(val in tbs_cats for val in local_cats))

        os.remove('test_product_sheet.csv')

        # test XLSX import, replace codes with new categories (update)
        cn8_codes = self.sa_session.query(WeCN8).limit(4).all()
        set_date(date(2015, 11, 1))
        headers = ['sku','name','description','is_b2b','cn8','weight_per_unit','length_per_unit','height_per_unit',
                   'width_per_unit','diagonal','weight_cat1','weight_cat2','weight_cat3','weight_cat4','weight_documentation',
                   'weight_batteries','weight_package','tbs-4','tbs-%s' % tbses[0].id, 'tbs-%s' % tbses[1].id, 'tbs-%s' % tbses[2].id,
                   'tbs-%s' % tbses[3].id, 'tbs-9','tbs-10','tbs-11','tbs-12']
        row1 = ['210-344','SKU1','','B2B',cn8_codes[0].code,'1.32','','999.999','','1000','','','','','','','','','11020','','','3.2','','','','']
        row2 = ['210-344','SKU1','','B2C',cn8_codes[0].code,'1.32','','999.999','','1000','','','','','','','','','11011','','','4.2','','','','']
        row3 = ['210-296','SKU3','','',cn8_codes[1].code,'3.837','','','','','','','','','','','','','','4.1.2','','','','','','']
        row4 = ['210-30','SKU4','','',cn8_codes[2].code,'2.283','','','','','','','','','','','','','','4.1.3','','','','','','']
        row5 = ['BG2036','SKU5','','',cn8_codes[3].code,'0.109','17.5','5.2','','','','','','','','','','','','','2/C06','','','','','']

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        sheet1.append(row5)
        book.save('test_product_sheet_xlsx.xlsx')

        local_cats = ['11020', '3.2', '11011', '4.2']

        # xlsx importer
        importer = HqppProductXlsxImporter(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        rv, failures = importer.import_file('test_product_sheet_xlsx.xlsx', self.controller,
                             hqpp_id=member.id, regulation_periods=d)
        self.assertEqual(len(failures), 0)
        # get product for sku
        product = self.sa_session.query(WeHqppProduct)\
            .filter(and_(WeHqppProduct.sku == '210-344',
                         WeHqppProduct.hqpp == member)).one()

        # get linked tbs category
        links = self.sa_session.query(WeHqppRegulationPeriodProduct)\
            .filter(WeHqppRegulationPeriodProduct.product_id == product.id).all()
        self.assertEqual(len(links), 4)

        tbs_cats = [c.tbs_regulation_period_product.tbs_product_level.label for c in links]
        self.assertTrue(all(val in tbs_cats for val in local_cats))
        os.remove('test_product_sheet_xlsx.xlsx')

    def test_import_product_sheet_multi_failures(self):
        set_date(date(2015, 5, 5))
        f_name, member, tbses, rps = self.create_product_sheet_csv_file()

        d = {
            'tbs-%s' % tbses[0].id: 'fail',
            'tbs-%s' % tbses[1].id: rps[1].id,
            'tbs-%s' % tbses[2].id: rps[2].id,
            'tbs-%s' % tbses[3].id: rps[3].id
        }

        os.remove('test_product_sheet.csv')

        # test XLSX import, replace codes with new categories (update)
        cn8_codes = self.sa_session.query(WeCN8).limit(4).all()
        set_date(date(2015, 11, 1))
        headers = ['sku','name','description','is_b2b','cn8','weight_per_unit','length_per_unit','height_per_unit',
                   'width_per_unit','diagonal','weight_cat1','weight_cat2','weight_cat3','weight_cat4','weight_documentation',
                   'weight_batteries','weight_package','tbs-4','tbs-%s' % tbses[0].id, 'tbs-%s' % tbses[1].id, 'tbs-%s' % tbses[2].id,
                   'tbs-%s' % tbses[3].id, 'tbs-9','tbs-10','tbs-11','tbs-12']
        row1 = ['210-344','SKU1','','B2B',cn8_codes[0].code,'1.32','','999.999','','1000','','','','','','','','','11020','','','fail','','','','']
        row2 = ['210-344','SKU1','','B2C',cn8_codes[0].code,'1.32','','999.999','','1000','','','','','','','','','','','','4.2','','','','']
        row3 = ['210-296','SKU3','','',cn8_codes[1].code,'3.837','','','','','','','','','','','','','','4.1.2','','','','','','']
        row4 = ['210-30','SKU4','','',cn8_codes[2].code,'2.283','','','','','','','','','','','','','','4.1.3','','','','','','']
        row5 = ['BG2036','SKU5','','',cn8_codes[3].code,'0.109','17.5','5.2','','','','','','','','','','','','','2/C06','','','','','']

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        sheet1.append(row5)
        book.save('test_product_sheet_xlsx.xlsx')

        # xlsx importer
        importer = HqppProductXlsxImporter(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        rv, failures = importer.import_file('test_product_sheet_xlsx.xlsx', self.controller,
                             hqpp_id=member.id, regulation_periods=d)
        self.assertEqual(len(failures), 1)
        fail1, fail2 = failures[0][len(failures[0])-1].split(' | ')
        self.assertTrue(fail1.startswith('Regulation period with ID fail'))
        self.assertTrue(fail2.startswith('There is no such category with label "fail"'))

    def test_export_product_sheet(self):
        # create tbses, hqpp, lpps, product levels
        set_date(date(2015, 5, 5))

        cn8s = self.sa_session.query(WeCN8).limit(2).all()
        member, lpps, tbses = self.create_lpps(self.get_vat_numbers(), 3)
        tbs1 = tbses[0]
        tbs2 = tbses[1]
        tbs3 = tbses[2]
        # tbs1 categories
        pl11, tbs = self.create_tbs_product_level(self._operator_account.id, tbs1, product_label='11010')
        pl12, tbs = self.create_tbs_product_level(self._operator_account.id, tbs1, product_label='11020')
        pl13, tbs = self.create_tbs_product_level(self._operator_account.id, tbs1, product_label='11030')
        pl14, tbs = self.create_tbs_product_level(self._operator_account.id, tbs1, product_label='11040')

        # tbs2 categories
        pl21, tbs = self.create_tbs_product_level(self._operator_account.id, tbs2, product_label='1.2')

        # tbs3 categories
        pl31, tbs = self.create_tbs_product_level(self._operator_account.id, tbs3, product_label='III')

        # create regulation periods
        rp1 = self.create_tbs_regulation_period(tbs1.id, self._uniquestring('RP_'), 2015, 1,
                                          [{'id': pl11.id, 'fee': Decimal('2.22')},
                                           {'id': pl12.id, 'fee': Decimal('0.03')}])
        rp2 = self.create_tbs_regulation_period(tbs1.id, self._uniquestring('RP_'), 2016, 1,
                                          [{'id': pl13.id, 'fee': Decimal('2.22')}])

        rp3 = self.create_tbs_regulation_period(tbs2.id, self._uniquestring('RP_'), 2015, 1,
                                          [{'id': pl21.id, 'fee': Decimal('2.22')}])

        rp4 = self.create_tbs_regulation_period(tbs3.id, self._uniquestring('RP_'), 2015, 1,
                                          [{'id': pl31.id, 'fee': Decimal('2.22')}])

        # save product1
        # selecting tbs product level which is not part of current regulation period should raise exception
        state = self.get_product_state(member.id, cn8s[0].code, pl13.id)
        self.assertRaises(PublicUserException, self.controller.save_or_update_hqpp_product, state['product_id'], state)
        state = self.get_product_state(member.id, cn8s[0].code, [pl11.id, pl21.id])
        product1 = self.controller.save_or_update_hqpp_product(state['product_id'], state)

        # save product 2
        state = self.get_product_state(member.id, cn8s[1].code, [pl12.id, pl31.id], sku='TEST123')
        product2 = self.controller.save_or_update_hqpp_product(state['product_id'], state)

        # export to file
        _file_name = self.controller.export_hqpp_product_sheet(member.id)

        from openpyxl.reader.excel import load_workbook
        wb = load_workbook(filename=_file_name, use_iterators=True)
        sheet = wb.get_active_sheet()
        row_count = 0
        for row in sheet.iter_rows():
            row_count += 1
            if row_count == 3:
                self.assertEqual(int(row[15].internal_value), cn8s[1].code)
                self.assertEqual(row[17].internal_value, pl12.label)
            # last row is with product1
            if row_count == 4:
                self.assertEqual(int(row[15].internal_value), cn8s[0].code)
                self.assertEqual(row[17].internal_value, pl11.label)
                self.assertEqual(row[18].internal_value, pl21.label)

        self.assertEqual(4, row_count)

    def test_export_product_sheet_no_products(self):
        # create tbses, hqpp, lpps
        set_date(date(2015, 5, 5))

        member, lpps, tbses = self.create_lpps(self.get_vat_numbers(), 3)

        # export to file
        _file_name = self.controller.export_hqpp_product_sheet(member.id)

        from openpyxl.reader.excel import load_workbook
        wb = load_workbook(filename=_file_name, use_iterators=True)
        sheet = wb.get_active_sheet()
        row_count = 0
        for row in sheet.iter_rows():
            row_count += 1
            if row_count == 1:
                vals = [t.internal_value for t in row[17:]]
                self.assertEqual(len(row[17:]), 3)
                self.assertTrue(all(t in vals for t in ['tbs-%s' % tbses[0].id,
                                                            'tbs-%s' % tbses[1].id,
                                                            'tbs-%s' % tbses[2].id]))

    def test_product_sheet_background_import(self):
        set_date(date(2015, 5, 5))
        f_name, member, tbses, rps = self.create_product_sheet_csv_file()
        # create dict with tbs_id: regulation_period_id
        d = {
            'tbs-%s' % tbses[0].id: rps[0].id,
            'tbs-%s' % tbses[1].id: rps[1].id,
            'tbs-%s' % tbses[2].id: rps[2].id,
            'tbs-%s' % tbses[3].id: rps[3].id
        }

        f = open(f_name, 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file(f_name, data)

        # xlsx importer
        importer = HqppProductAbstractImporter().get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 2
        importer.import_file(myfile, self.controller, hqpp_id=member.id, regulation_periods=d)

        # get the saved file
        f = self.sa_session.query(ImportFile).join(ImportFileType)\
            .filter(and_(ImportFileType.name == importer.IMPORT_FILE_TYPE, ImportFile.status == ImportFile.STATUS_NEW))\
            .order_by(ImportFile.last_updated.desc()).first()
        self.assertEqual(f.file_name, f_name)
        settings = simplejson.loads(f.settings)
        self.assertEqual(settings['regulation_periods'], d)
        import time
        time.sleep(1)

        # Trigger task
        HqppProductSheetTask().execute()

        products = self.sa_session.query(WeHqppProduct).filter(WeHqppProduct.hqpp_id == member.id).all()
        self.assertEqual(len(products), 4)
        product_skus = [p.name for p in products]
        self.assertTrue(all(n in product_skus for n in ['SKU1', 'SKU3', 'SKU4', 'SKU5']))

        # check notification is created
        nc = NotificationController(self.sa_session, account_id=self._operator_account.id)
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 1)
        self.assertEqual(notif[0].callback, '{"js_module": "we_hqpp_product_manager"}')

    def test_import_tbs_price_sheet(self):

        # import for new regulation period - create period
        # import for old period - update records
        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Fee']
        row1 = ['', '1', 'Large Household', 'Quantity', Decimal('0.08')]
        row2 = ['1', '1.1', 'Fridges', '', Decimal('0.08')]
        row3 = ['', '2', 'Small Household', 'Quantity', Decimal('10.08')]
        row4 = ['2', '2.1', 'Microwave', 'fixed annual fee', Decimal('0.88')]
        row5 = ['', '3', 'TV', 'Weight', Decimal('0.88')]

        import csv
        with open('test_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)
            wr.writerow(row5)

        f = open('test_import_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_import_categories.csv', data)

        tbs = self._create_tbs(self._operator_account.id)
        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        row_count = importer.get_row_count('test_import_categories.csv')
        self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                               None, 'Test RP csv', 2015, 1, 1)

        level = self.sa_session.query(WeTbsProductLevel).filter(and_(WeTbsProductLevel.tbs == tbs,
                                                                     WeTbsProductLevel.label == '2.1',
                                                                     WeTbsProductLevel.position == 1)).first()
        self.assertEqual(level.description, 'Microwave')
        self.assertEqual(level.invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_FIXED_ANNUAL_FEE)

        # check regulation period is created
        rp = self.controller.get_regulation_periods(tbs.id, 'Test RP csv').pop()
        self.assertEqual(rp.start_date, date(2015, 1, 1))

        trpp = self.sa_session.query(WeTbsRegulationPeriodProduct)\
            .filter(and_(WeTbsRegulationPeriodProduct.tbs_product_level == level,
                         WeTbsRegulationPeriodProduct.tbs_regulation_period == rp)).first()
        self.assertEqual(trpp.fee, Decimal('0.88000'))

        os.remove('test_import_categories.csv')

        # update the regulation period
        row1[-1] = Decimal('22.22')
        row2[-1] = Decimal('22.22')
        row3[-1] = Decimal('22.22')
        row4[-1] = Decimal('22.22')
        row5[-1] = Decimal('22.22')
        with open('test_update_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)
            wr.writerow(row5)

        f = open('test_update_import_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_update_import_categories.csv', data)

        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                               rp.id, 'Test RP update csv', 2015, 1)

        level = self.sa_session.query(WeTbsRegulationPeriodProduct)\
            .filter(and_(WeTbsRegulationPeriodProduct.tbs_product_level_id == level.id,
                         WeTbsRegulationPeriodProduct.tbs_regulation_period_id == rp.id)).first()
        self.assertEqual(level.fee, Decimal('22.22'))

        # test xlsx import
        tbs1 = self._create_tbs(self._operator_account.id)

        file_path = os.path.join(os.path.dirname(__file__), os.pardir, 'files', 'local_category_test_import.xlsx')
        importer = TbsPriceSheetXlsxImporter(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 100
        row_count = importer.get_row_count(file_path)
        self.controller.import_tbs_price_sheet(file_path, importer, tbs1.id, row_count, None,
                                               'Test RP xlsx', 2015, 3, 1)

        # check that all categories are imported
        levels = self.sa_session.query(WeTbsProductLevel).filter(WeTbsProductLevel.tbs == tbs1).all()
        self.assertEqual(len(levels), 24)

        rp = self.controller.get_regulation_periods(tbs1.id, 'Test RP xlsx').pop()
        self.assertEqual(rp.start_date, date(2015, 3, 1))

    def test_import_tbs_price_sheet_no_fee_on_main_category(self):

        # import for new regulation period - create period
        # import for old period - update records
        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Fee']
        row1 = ['', '1', 'Large Household', 'Quantity', '']
        row2 = ['1', '1.1', 'Fridges', '', Decimal('0.08')]
        row3 = ['', '2', 'Small Household', 'Quantity', '']
        row4 = ['2', '2.1', 'Microwave', 'fixed annual fee', Decimal('0.88')]
        row5 = ['', '3', 'TV', 'Weight', Decimal('0.88')]

        import csv
        with open('test_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)
            wr.writerow(row5)

        f = open('test_import_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_import_categories.csv', data)

        tbs = self._create_tbs(self._operator_account.id)
        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        row_count = importer.get_row_count('test_import_categories.csv')
        self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                               None, 'Test RP csv', 2015, 1, 1)

        level = self.sa_session.query(WeTbsProductLevel).filter(and_(WeTbsProductLevel.tbs == tbs,
                                                                     WeTbsProductLevel.label == '2.1',
                                                                     WeTbsProductLevel.position == 1)).first()
        self.assertEqual(level.description, 'Microwave')
        self.assertEqual(level.invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_FIXED_ANNUAL_FEE)

        # check regulation period is created
        rp = self.controller.get_regulation_periods(tbs.id, 'Test RP csv').pop()
        self.assertEqual(rp.start_date, date(2015, 1, 1))
        self.assertEqual(len(rp.period_products), 5)
        available = self.controller.get_tbs_product_levels_for_period(tbs.id, date(2015, 1, 1), include_null_fees=False)
        self.assertEqual(len(available), 3)

    def test_import_tbs_price_sheet_different_description(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs,
                                                            product_label='11010', description='YOLO')

        # import for new regulation period - create period
        # import for old period - update records
        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Fee']
        row1 = ['', '11010', 'Large Household', 'Quantity', '']
        row2 = ['11010', '1.1', 'Fridges', '', Decimal('0.08')]
        row3 = ['', '2', 'Small Household', 'Quantity', '']
        row4 = ['2', '2.1', 'Microwave', 'fixed annual fee', Decimal('0.88')]
        row5 = ['', '3', 'TV', 'Weight', Decimal('0.88')]

        import csv
        with open('test_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)
            wr.writerow(row5)

        f = open('test_import_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_import_categories.csv', data)

        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        row_count = importer.get_row_count('test_import_categories.csv')
        self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                               None, 'Test RP csv', 2015, 1, 1)

        level = self.sa_session.query(WeTbsProductLevel).filter(and_(WeTbsProductLevel.tbs == tbs,
                                                                     WeTbsProductLevel.label == '2.1',
                                                                     WeTbsProductLevel.position == 1)).first()
        self.assertEqual(level.description, 'Microwave')
        self.assertEqual(level.invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_FIXED_ANNUAL_FEE)

        # check regulation period is created
        rp = self.controller.get_regulation_periods(tbs.id, 'Test RP csv').pop()
        self.assertEqual(rp.start_date, date(2015, 1, 1))
        self.assertEqual(len(rp.period_products), 5)
        available = self.controller.get_tbs_product_levels_for_period(tbs.id, date(2015, 1, 1), include_null_fees=False)
        self.assertEqual(len(available), 3)
        self.assertTrue(product_level1 not in available)

    def test_import_tbs_price_sheet_unarchive_product(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs,
                                                            product_label='11010', description='Large Household')

        product_level1.archived = True
        self.sa_session.add(product_level1)
        self.sa_session.flush()

        # import for new regulation period - create period
        # import for old period - update records
        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Fee']
        row1 = ['', '11010', 'Large Household', 'Quantity', '']
        row2 = ['11010', '1.1', 'Fridges', '', Decimal('0.08')]
        row3 = ['', '2', 'Small Household', 'Quantity', '']
        row4 = ['2', '2.1', 'Microwave', 'fixed annual fee', Decimal('0.88')]
        row5 = ['', '3', 'TV', 'Weight', Decimal('0.88')]

        import csv
        with open('test_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)
            wr.writerow(row5)

        f = open('test_import_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_import_categories.csv', data)

        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        row_count = importer.get_row_count('test_import_categories.csv')
        self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                               None, 'Test RP csv', 2015, 1, 1)
        self.sa_session.expire(product_level1)
        self.assertFalse(product_level1.archived)

    def test_import_tbs_price_sheet_main_category_not_in_file(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs,
                                                            product_label='11010', description='YOLO')

        # import for new regulation period - create period
        # import for old period - update records
        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Fee']
        row1 = ['', '11010', 'Large Household', 'Quantity', '']
        row2 = ['11010', '1.1', 'Fridges', '', Decimal('0.08')]
        row3 = ['', '2', 'Small Household', 'Quantity', '']
        row4 = ['4', '2.1', 'Microwave', 'fixed annual fee', Decimal('0.88')]  # ERROR, 4 not in file
        row5 = ['3', '3.1', 'TV Remote', 'Quantity', Decimal('0.88')]  # ERROR, 3 comes AFTER this row.
        row6 = ['', '3', 'TV', 'Weight', Decimal('0.88')]

        import csv
        with open('test_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)
            wr.writerow(row5)
            wr.writerow(row6)

        f = open('test_import_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_import_categories.csv', data)

        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        row_count = importer.get_row_count('test_import_categories.csv')
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        failures = self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                               None, 'Test RP csv', 2015, 1, 1)

        self.assertEqual(len(failures), 2)

    def test_import_tbs_price_sheet_empty_fee_on_subcategory(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs,
                                                            product_label='11010', description='YOLO')

        # import for new regulation period - create period
        # import for old period - update records
        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Fee']
        row1 = ['', '11010', 'Large Household', 'Quantity', '']
        row2 = ['11010', '1.1', 'Fridges', '', '  ']

        import csv
        with open('test_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)

        f = open('test_import_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_import_categories.csv', data)

        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        row_count = importer.get_row_count('test_import_categories.csv')
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        failures = self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                               None, 'Test RP csv', 2015, 1, 1)

        self.assertEqual(len(failures), 1)
        self.assertTrue(failures[0][-1].startswith("Fee should be provided for this category"))

    def test_background_import_tbs_product_level(self):

        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Current fee']
        row1 = ['', '1', 'Large Household', 'Quantity', Decimal('0.08')]
        row2 = ['1', '1.1', 'Fridges', '', '']
        row3 = ['', '2', 'Small Household', 'Quantity', Decimal('10.08')]
        row4 = ['2', '2.1', 'Microwave', 'Weight', Decimal('0.88')]

        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet1 = book.get_active_sheet()
        sheet1.append(headers)
        sheet1.append(row1)
        sheet1.append(row2)
        sheet1.append(row3)
        sheet1.append(row4)
        book.save('test_import_categories.xlsx')

        tbs = self._create_tbs(self._operator_account.id)
        f = open('test_import_categories.xlsx', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_import_categories.xlsx', data)

        # xlsx importer
        importer = TbsPriceSheetXlsxImporter(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 2
        row_count = importer.get_row_count('test_import_categories.xlsx')
        importer.import_file(myfile, self.controller, tbs_id=tbs.id, regulation_period_id=None,
                             regulation_label='Test BG RP xlsx', regulation_year=2015,
                             regulation_month=3, weight_settings=1, row_count=row_count, account_id=self._operator_account.id)

        # get the saved file
        f = self.sa_session.query(ImportFile).join(ImportFileType)\
            .filter(and_(ImportFileType.name == importer.IMPORT_FILE_TYPE, ImportFile.status == ImportFile.STATUS_NEW))\
            .order_by(ImportFile.last_updated.desc()).first()
        self.assertEqual(f.file_name, 'test_import_categories.xlsx')

        import time
        time.sleep(1)

        # Trigger task
        TbsPriceSheetImportTask().execute()
        levels = self.controller.get_tbs_product_levels(None, tbs_id=tbs.id)
        self.assertEqual(len(levels), 4)
        self.assertEqual(levels[1].label, '1.1')
        self.assertEqual(levels[1].description, 'Fridges')
        self.assertEqual(levels[1].invoice_basis, WeTbsProductLevel.INVOICE_BASED_ON_QUANTITY)

        rp = self.controller.get_regulation_periods(tbs.id, 'Test BG RP xlsx', date(2015, 3, 1))
        rpp = levels[1].get_current_regulation_period_product(rp.id)
        self.assertEqual(rpp.fee, Decimal('0.08')) # inherited fee from parent category

        # check notification is created
        nc = NotificationController(self.sa_session, account_id=self._operator_account.id)
        notif = nc.get_unread_notifications()
        self.assertEqual(len(notif), 1)
        self.assertEqual(notif[0].callback, '{"js_module": "we_tbs_regulation_period_compare", '
                                            '"parameters": [%s]}' % rp.id)

    def test_tbs_product_level_import_failure_lines(self):
        headers = ['Main category label', 'Category label', 'Category description', 'Basis for invoice', 'Current fee']
        row1 = ['', '1', 'Large Household', 'Quantity', '']
        row2 = ['1', '1.1', 'Fridges', '', Decimal('0.88')]
        row3 = ['3', '2', 'Small Household', 'Quantity', ''] # failure - no such main category
        row4 = ['2', '2.1', 'Microwave', 'Weight', ''] # failure - main category was not imported

        import csv
        with open('test_import_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)

        tbs = self._create_tbs(self._operator_account.id)
        importer = TbsPriceSheetCsvImporter(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 10
        row_count = importer.get_row_count('test_import_categories.csv')
        failures = self.controller.import_tbs_price_sheet('test_import_categories.csv', importer,
                                                          tbs.id, row_count, None,
                                                          'Test fails RP csv', 2015, 3, weight_settings=1)
        self.assertEqual(len(failures), 2)

    def test_import_cn8_to_product_level(self):
        # create csv with cn8 and tbs product levels which exist in the database
        codes = self.sa_session.query(WeCN8).limit(5).all()
        headers = ['CN8', 'Local category', 'B2B/B2C']
        row1 = [codes[0].code, '11010', 'b2b']
        row2 = [codes[1].code, '11011', 'b2c']
        row3 = [codes[2].code, '11020', '']
        row4 = [codes[3].code, 11030, 'b2b']
        row5 = [codes[4].code, '45030', 'b2b'] # failure - local cat missing

        import csv
        with open('test_cn8_to_local_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)
            wr.writerow(row5)

        tbs = self._create_tbs(self._operator_account.id)
        pl1, bitbucket = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11010')
        pl2, bitbucket = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11011')
        pl3, bitbucket = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11020')
        pl4, bitbucket = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11030')

        rp = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        },{
            'id': pl4.id,
            'fee': Decimal('0.52')
        }])

        f = open('test_cn8_to_local_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_cn8_to_local_categories.csv', data)

        importer = CN8ToTbsProductLevelAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 2
        self.controller.import_cn8_to_tbs_product_level(myfile, importer, tbs_id=tbs.id, regulation_period_id=rp.id)

        # Trigger task
        CN8ToTbsProductLevelTask().execute()

        f = self.sa_session.query(ImportFile).join(ImportFileType)\
            .filter(and_(ImportFileType.name == importer.IMPORT_FILE_TYPE, ImportFile.status == ImportFile.STATUS_SUCCESS))\
            .order_by(ImportFile.last_updated.desc()).first()
        self.assertEqual(f.file_name, 'test_cn8_to_local_categories.csv')
        self.assertEqual(f.settings, '{"regulation_period_id": %s, "tbs_id": %s}' % (rp.id, tbs.id))

        file_path = importer.get_file_path(f.failure_file)
        f = open(file_path, 'r')
        n = 0
        row = ''
        for line in f:
            n += 1
            if n == 2:
                row = line
        self.assertEqual(n, 2)
        self.assertNotEqual(row.find('45030'), -1)
        f.close()

        suggested = self.controller.get_suggested_tbs_product_levels(tbs.id, codes[0].code, date(2015, 6, 1))
        self.assertEqual(len(suggested), 1)
        self.assertEqual(suggested[0], pl1)

    def test_import_cn8_to_product_level_duplicates(self):
        """
        Duplicates are not allowed.
        :return:
        """
        # create csv with cn8 and tbs product levels which exist in the database
        codes = self.sa_session.query(WeCN8).limit(3).all()
        headers = ['CN8', 'Local category', 'B2B/B2C']
        row1 = [codes[0].code, '11010', 'b2b']
        row2 = [codes[1].code, '11011', 'b2c']
        row3 = [codes[2].code, '11020', 'b2b']
        row4 = [codes[2].code, '11020', 'b2b']

        import csv
        with open('test_cn8_to_local_categories.csv', 'wb') as f:
            wr = csv.writer(f, dialect=csv.excel, delimiter=',')
            wr.writerow(headers)
            wr.writerow(row1)
            wr.writerow(row2)
            wr.writerow(row3)
            wr.writerow(row4)

        tbs = self._create_tbs(self._operator_account.id)
        pl1, bitbucket = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11010')
        pl2, bitbucket = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11011')
        pl3, bitbucket = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='11020')

        rp = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])

        f = open('test_cn8_to_local_categories.csv', 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_cn8_to_local_categories.csv', data)

        importer = CN8ToTbsProductLevelAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 100
        failures = self.controller.import_cn8_to_tbs_product_level(myfile, importer, tbs.id, rp.id)

        self.assertEqual(len(failures), 1)
        self.assertTrue(failures[0][-1].startswith("There are duplicate"))

    def test_cn8_to_weee_category_import(self):
        file_path = os.path.join(os.path.dirname(__file__), os.pardir, 'files', 'test_cn8.xlsx')
        f = open(file_path, 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file('test_cn8_to_weee_category.xlsx', data)

        # xlsx importer
        importer = CN8AbstractImporter().get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 2
        importer.import_file(myfile, self.controller)

        # get the saved file
        f = self.sa_session.query(ImportFile).join(ImportFileType)\
            .filter(and_(ImportFileType.name == importer.IMPORT_FILE_TYPE, ImportFile.status == ImportFile.STATUS_NEW))\
            .order_by(ImportFile.last_updated.desc()).first()
        self.assertEqual(f.file_name, 'test_cn8_to_weee_category.xlsx')

        import time
        time.sleep(1)

        # Trigger task
        CN8ToWeeeCategoryTask().execute()
        code = self.sa_session.query(WeCN8).filter(WeCN8.code == 77777777).one()

        self.assertEqual(code.unu_code.code, '0666')
        self.assertEqual(code.weee_product_level.label, 'EEE-1.03.01')

    def test_regulation_period_creation(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        rp = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP', 2015, 1, 1, [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        },{
            'id': product_level2.id,
            'fee': Decimal('33.33')
        },{
            'id': product_level3.id,
            'fee': Decimal('12.45')
        }])

        pp = [p.tbs_product_level for p in rp.period_products]

        self.assertTrue(product_level1 in pp)
        self.assertTrue(product_level2 in pp)
        self.assertTrue(product_level3 in pp)

    def test_regulation_period_with_negative_fee(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        categories = [{
            'id': product_level1.id,
            'fee': Decimal('-20.12')
        },{
            'id': product_level2.id,
            'fee': Decimal('-33.33')
        }]
        self.assertRaises(PublicUserException, self.controller.save_or_update_regulation_period,
                          None, tbs.id, 'Test RP', 2015, 5, 1, categories)

        tbs.set_config('we.tbs.use.negative.fee', True)

        rp = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP', 2015, 5, 1, categories)
        pp = [p.tbs_product_level for p in rp.period_products]

        self.assertTrue(product_level1 in pp)
        self.assertTrue(product_level2 in pp)

    def test_regulation_period_incorrect_product(self):
        tbs = self._create_tbs(self._operator_account.id)
        tbs1 = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        self.assertRaises(PublicUserException, self.controller.save_or_update_regulation_period, None, tbs1.id, 'Test RP', 2015, 1, 1, [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        }])

    def test_regulation_period_update_for_past_date(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        rp = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP', 2015, 1, 1, [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        }])

        self.assertRaises(PublicUserException, self.controller.save_or_update_regulation_period, rp.id, tbs.id, 'Test RP', 2015, 10, 1, [{
            'id': product_level1.id,
            'fee': Decimal('11.12')
        }])

    def test_regulation_period_same_category(self):
        file_path = os.path.join(os.path.dirname(__file__), os.pardir, 'files', 'regulation_period_same_category.txt')
        f = open(file_path, 'r')
        data = f.read()
        f.close()
        myfile = self._create_uploaded_file(file_path, data)

        tbs = self._create_tbs(self._operator_account.id)

        importer = TbsPriceSheetAbstractImporter()\
            .get_importer(myfile)(self.sa_session, account_id=self._operator_account.id)
        importer.MAX_RECORDS_FOR_INSTANT_IMPORT = 100
        row_count = importer.get_row_count(file_path)
        failures = self.controller.import_tbs_price_sheet(myfile, importer, tbs.id, row_count,
                                                          None, 'Test RP csv', 2015, 1, 1)

        self.assertEqual(len(failures), 1)
        self.assertTrue(failures[0][-1].startswith("You cannot have two categories"))

    def _setup_regulation_period_compare(self):
        cn8 = self.sa_session.query(WeCN8).first()

        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='1.1.1')
        product_level2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        rp = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP', 2015, 1, 1, [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        },{
            'id': product_level2.id,
            'fee': Decimal('33.33')
        },{
            'id': product_level3.id,
            'fee': Decimal('12.45')
        }])

        comp = self.controller.compare_regulation_period_with_previous(rp.id)
        self.assertEqual(comp, None)  # there is no previous period

        rpl1 = rp.period_products[0]
        rpl2 = rp.period_products[1]
        rpl3 = rp.period_products[2]

        # add a few producers
        member1_products = self._add_member_products(tbs, [rpl1, rpl3], self._operator_account.id)
        member2_products = self._add_member_products(tbs, [rpl1, rpl2], self._operator_account.id)
        member3_products = self._add_member_products(tbs, [rpl1, rpl2, rpl3], self._operator_account.id)
        member4_products = self._add_member_products(tbs, [rpl2, rpl3], self._operator_account.id)
        member5_products = self._add_member_products(tbs, [rpl3], self._operator_account.id)

        # make a new period
        product_level4, tbs = self.create_tbs_product_level(self._operator_account.id, tbs, product_label='1.1.1', description='Different than the previous')
        product_level5, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        rp2 = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP 1', 2015, 6, 1, [{
            'id': product_level4.id,
            'fee': Decimal('20.12')
        },{
            'id': product_level5.id,
            'fee': Decimal('33.33')
        },{
            'id': product_level3.id,
            'fee': Decimal('12.45')
        }])

        comp = self.controller.compare_regulation_period_with_previous(rp2.id)
        self.assertEqual(comp['nochange'][0][0]['tbs_product_level_id'], product_level3.id)
        self.assertEqual(comp['removed'][0][0]['tbs_product_level_id'], product_level2.id)
        self.assertEqual(comp['modified'][0][0]['tbs_product_level_id'], product_level1.id)

        self.assertEqual(len(comp['affected']), 4)  # member 1-4 carry affected products
        self.assertEqual(len(comp['automatic']), 1)  # member 5 does not.

        return rp2, [member1_products, member2_products, member3_products, member4_products, member5_products]

    def _assert_new_period(self, member_products):

        m1 = member_products[0][0].product.hqpp
        m2 = member_products[1][0].product.hqpp
        m3 = member_products[2][0].product.hqpp
        m4 = member_products[3][0].product.hqpp
        m5 = member_products[4][0].product.hqpp

        # members 1-4 should have a notification now.
        for m in [m1, m2, m3, m4]:
            nc = WeNotificationController(self.sa_session, account_id=m.default_group.accounts[0].id)
            notification = nc.get_unread_notifications().pop()
            self.assertTrue(isinstance(notification, WeRegulationPeriodNotification))

        # member 5 should not have a notification.
        nc = WeNotificationController(self.sa_session, account_id=m5.default_group.accounts[0].id)
        self.assertEqual(len(nc.get_unread_notifications()), 0)

    def test_regulation_period_compare(self):
        rp2, member_products = self._setup_regulation_period_compare()
        self.controller.compare_regulation_period_with_previous(rp2.id, dry_run=False)
        self._assert_new_period(member_products)

    def test_regulation_period_task(self):
        rp2, member_products = self._setup_regulation_period_compare()
        set_date(date(2015,6,1))
        RegulationPeriodChangeTask().execute()
        self._assert_new_period(member_products)

    def test_regulation_period_remove_product_level(self):
        tbs = self._create_tbs(self._operator_account.id)
        product_level1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        product_level3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)

        pl_id = product_level3.id

        categories = [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        },{
            'id': product_level2.id,
            'fee': Decimal('33.33')
        },{
            'id': product_level3.id,
            'fee': Decimal('12.45')
        }]

        rp = self.controller.save_or_update_regulation_period(None, tbs.id, 'Test RP', 2015, 1, 1, categories)

        # update by removing one category - this should not delete the actual tbs product level
        categories = [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        },{
            'id': product_level2.id,
            'fee': Decimal('33.33')
        }]

        rp = self.controller.save_or_update_regulation_period(rp.id, tbs.id, 'Test RP', 2015, 1, 1, categories)

        # this should not raise an error
        pl3 = self.sa_session.query(WeTbsProductLevel).filter(WeTbsProductLevel.id == pl_id).one()

        # add producer products
        rpl2 = rp.period_products[1]
        self._add_member_products(tbs, [rpl2], self._operator_account.id)

        # check that the product level cannot be removed from the regulation period
        categories = [{
            'id': product_level1.id,
            'fee': Decimal('20.12')
        }]

        self.assertRaises(PublicUserException, self.controller.save_or_update_regulation_period,
                          rp.id, tbs.id, 'Test RP', 2015, 1, 1, categories)