# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

import suds
from datetime import date
from decimal import Decimal

from django.conf import settings

from basesoft.util.virtual_date import set_date

from we.declaration.controllers import WeDeclarationController
from we.order.controllers import WeDeclarationOrderController
from we.order.models import WeDeclarationOrder
from we.product.models import WeCN8
from we.tbs.models import WeTbsLicense
from we.test.unit.we_unittest import WeUnitTestCase


class TestCase(WeUnitTestCase):
    def setUp(self):
        self.connect_db()

        self._start_django()
        self.client = suds.client.Client('%swe.core/api/wsdl' % settings.BASE_URL, cache=None)
        self._operator_account = self._create_operator_account()
        set_date(date(2015, 1, 1))

    def _prepare_declarations(self, periodicity):
        controller = WeDeclarationController(self.sa_session, account_id=self._operator_account.id)

        member, lpp, tbs_product_level, tbs = self.create_lpp()

        tbs.set_config('we.tbs.declaration.period', periodicity)
        tbs.set_config('we.tbs.declaration.closing.day', 15)

        pl1, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl2, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        pl3, tbs = self.create_tbs_product_level(self._operator_account.id, tbs)
        rp1 = self.create_tbs_regulation_period(tbs.id, 'label', 2015, 1, [{
            'id': pl1.id,
            'fee': Decimal('12.345')
        },{
            'id': pl2.id,
            'fee': Decimal('1.002')
        },{
            'id': pl3.id,
            'fee': Decimal('0.56')
        }])

        set_date(date(2015, 4, 1))

        product = self.add_hqpp_product(member.id, pl1.id, '73211190', 'UC389458', period=rp1.start_date)
        product1 = self.add_hqpp_product(member.id, pl2.id, '84031010', 'WEE34958', period=rp1.start_date)
        product2 = self.add_hqpp_product(member.id, pl3.id, '84185090', 'XXX666', period=rp1.start_date)

        period_product = product.get_period_product(rp1)[0]
        period_product1 = product1.get_period_product(rp1)[0]
        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 2,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 5,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 2,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        declaration1 = controller.save_or_update_declaration(None, state)

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 3,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 12,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 8,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        declaration2 = controller.save_or_update_declaration(None, state)

        # Save a Declaration NOT in this Declaration period, it should not be included in the export.

        state = {
            'lpp_id': lpp.id,
            'year': 2015,
            'month': 5,
            'lines': [
                {
                    'line_id': '',
                    'product_id': product.id,
                    'sku': 'UC389458',
                    'quantity': 16,
                    'hqpp_period_product_id': period_product.id
                }, {
                    'line_id': None,
                    'product_id': product1.id,
                    'sku': 'WEE34958',
                    'quantity': 18,
                    'hqpp_period_product_id': period_product1.id
                }
            ]
        }
        declaration3 = controller.save_or_update_declaration(None, state)

        return tbs, (declaration1, declaration2, declaration3,), (period_product, period_product1,)

    def test_get_unexported_declaration_orders(self):

        tbs, declarations, products = self._prepare_declarations(WeTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        set_date(date(2015, 4, 15))

        tbs_account = self._create_active_account(tbs)

        # we do not have access to login
        self.assertRaises(suds.WebFault, self.client.service.login, tbs_account.name, 'test_password')
        # login is required for exported orders
        self.assertRaises(suds.WebFault, self.client.service.get_unexported_orders_for_merchant, tbs.id)

        self._attach_permissions_to_group(tbs_account.default_group, 'soap_login', 'we_view_declaration_order')
        self.client.service.login(tbs_account.name, 'test_password')

        rv = self.client.service.get_unexported_orders_for_merchant(tbs.id)
        self.assertEqual(len(rv), 0)

        wod = WeDeclarationOrderController(self.sa_session)
        orders = wod.process_declarations(tbs.id)
        self.assertEqual(len(orders), 1)
        order = orders.pop()
        member = order.client

        wod = WeDeclarationOrderController(self.sa_session, account_id=tbs_account.id)
        wod.assert_privileges_on_object(order, WeDeclarationOrder.PRIV_EXPORT)

        rv = self.client.service.get_unexported_orders_for_merchant(tbs.id)
        self.assertEqual(len(rv), 1)

        orders = rv.item
        self.assertEqual(len(orders), 1)
        order = orders.pop()

        self.assertEqual(str(order.total_amount_excluding_vat), '219.885')
        self.assertEqual(order.currency_code, 'EUR')
        self.assertEqual(order.current_payment_method_id, None)

        props = order.properties[0]

        self.assertEqual(len(order.lines[0]), 2)
        self.assertEqual(order.merchant_id, tbs.id)
        self.assertEqual(order.client_id, member.id)
        self.assertEqual(props[0].value_int, 2015)  # year
        self.assertEqual(props[1].value_int, None)  # month
        self.assertEqual(props[2].value_int, 1)  # quarter
        self.assertEqual(props[3].value_int, None)  # bi annual
        self.assertEqual(props[4].value_str, member.lpps[0].lpp_relation_id)

        self.assertEqual(order.lines[0][0]['vat_rate'], 0)
        self.assertEqual(order.lines[0][0]['quantity'], 17)
        self.assertEqual(str(order.lines[0][0]['properties'][0][0].value_decimal), '36.091')
        self.assertEqual(order.lines[0][0]['properties'][0][1].value_str, products[0].tbs_regulation_period_product.tbs_product_level.label)

    def test_get_declarations_by_cn8(self):
        tbs, declarations, products = self._prepare_declarations(WeTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        set_date(date(2015, 4, 15))

        tbs_account = self._create_active_account(tbs)

        wod = WeDeclarationOrderController(self.sa_session)
        orders = wod.process_declarations(tbs.id)
        self.assertEqual(len(orders), 1)

        self._attach_permissions_to_group(tbs_account.default_group, 'soap_login', 'we_view_declaration', 'we_view_declaration_order')
        self.client.service.login(tbs_account.name, 'test_password')

        rv = self.client.service.get_declarations_by_cn8(tbs.id, 2015, None, 1, None)
        decl = rv.item.pop()
        self.assertEqual(len(decl.lines[0]), 2)

        # first line
        self.assertEqual(decl.lines[0][0].cn8, '73211190')
        self.assertEqual(str(decl.lines[0][0].weight), '36.091')
        self.assertEqual(decl.lines[0][0].quantity, 17)

        # second line
        self.assertEqual(decl.lines[0][1].cn8, '84031010')
        self.assertEqual(str(decl.lines[0][1].weight), '21.23')
        self.assertEqual(decl.lines[0][1].quantity, 10)
